import os
import unittest

import ai2thor
from PIL import Image, ImageChops, ImageStat

from machine_common_sense.config_manager import (FloorPartitionConfig,
                                                 SceneConfiguration,
                                                 Vector2dInt, Vector3d)
from machine_common_sense.plotter import (Ramp, SceneAsset, SceneBounds,
                                          SceneCoord, TopDownPlotter,
                                          XZHeading)

test_path = os.path.dirname(__file__)
resources_path = os.path.join(test_path, 'resources')


class TestSceneCoord(unittest.TestCase):

    def test_add(self):
        sc1 = SceneCoord(x=-1.0, y=0, z=2.5)
        sc2 = SceneCoord(x=1.0, y=0, z=-2.5)
        sc_add = sc1 + sc2
        self.assertAlmostEqual(sc_add.x, 0)
        self.assertAlmostEqual(sc_add.z, 0)

    def test_subtract(self):
        sc1 = SceneCoord(x=-1.0, y=0, z=2.5)
        sc2 = SceneCoord(x=1.0, y=0, z=-2.5)

        sc_sub = sc1 - sc2
        self.assertAlmostEqual(sc_sub.x, -2.0)
        self.assertAlmostEqual(sc_sub.z, 5.0)

        # order matters
        sc_sub = sc2 - sc1
        self.assertAlmostEqual(sc_sub.x, 2.0)
        self.assertAlmostEqual(sc_sub.z, -5.0)

    def test_midpoint(self):
        '''Gets the midpoint between the two points'''
        sc1 = SceneCoord(x=-1.0, y=0, z=2.5)
        sc2 = SceneCoord(x=1.0, y=0, z=-2.5)

        sc_sub = sc1 | sc2
        self.assertAlmostEqual(sc_sub.x, 0.0)
        self.assertAlmostEqual(sc_sub.z, 0.0)

        # order shouldn't matter
        sc_sub = sc2 | sc1
        self.assertAlmostEqual(sc_sub.x, 0.0)
        self.assertAlmostEqual(sc_sub.z, 0.0)


class TestRamp(unittest.TestCase):

    bounds = [SceneCoord(**pt) for pt in [
        {'x': 1.5, 'y': 0, 'z': 2},
        {'x': 2.5, 'y': 0, 'z': 2},
        {'x': 2.5, 'y': 0, 'z': -2},
        {'x': 1.5, 'y': 0, 'z': -2},
        {'x': 1.5, 'y': 2, 'z': 2},
        {'x': 2.5, 'y': 2, 'z': 2},
        {'x': 2.5, 'y': 2, 'z': -2},
        {'x': 1.5, 'y': 2, 'z': -2}
    ]]

    def test_ramp_arrow(self):
        ramp = Ramp(
            held=False,
            visible=True,
            uuid="ramp-test",
            color="blue",
            bounds=SceneBounds(self.bounds))
        arrow = ramp.arrow
        self.assertTrue(len(arrow.floor), 2)
        self.assertIsInstance(arrow.floor, list)
        self.assertIsInstance(arrow.floor[0], SceneCoord)

        self.assertAlmostEqual(arrow.floor[0].x, self.bounds[3].x)
        self.assertAlmostEqual(arrow.floor[0].z, self.bounds[3].z)

        self.assertAlmostEqual(arrow.floor[1].x, self.bounds[2].x)
        self.assertAlmostEqual(arrow.floor[1].z, self.bounds[2].z)

        self.assertIsInstance(arrow.peak, SceneCoord)
        self.assertAlmostEqual(arrow.peak.x, 2.0)
        self.assertAlmostEqual(arrow.peak.z, 2.0)


class TestTopDownPlotter(unittest.TestCase):

    def setUp(self):
        scene_config = SceneConfiguration(
            name="test", room_dimensions=Vector3d(
                x=10, y=4, z=10))
        self.plotter = TopDownPlotter(
            team="test",
            scene_config=scene_config
        )

    def test_convert_color_empty(self):
        color = self.plotter._convert_color('')
        self.assertEqual(color, "ivory")

    def test_convert_color_none(self):
        color = self.plotter._convert_color(None)
        self.assertEqual(color, "ivory")

    def test_convert_color_black(self):
        color = self.plotter._convert_color('black')
        self.assertEqual(color, "ivory")

    def test_convert_color(self):
        color = self.plotter._convert_color('red')
        self.assertEqual(color, "red")

    def test_plot_image_size(self):
        metadata = {
            'screenWidth': 600,
            'screenHeight': 400,
            'objects': [],
            'structuralObjects': [],
            'agent': {
                'position': {
                    'x': 0.0,
                    'y': 0.0,
                    'z': 0.0
                },
                'rotation': {
                    'x': 0.0,
                    'y': 0.0,
                    'z': 0.0
                }
            }}
        scene_event = ai2thor.server.Event(metadata=metadata)
        img = self.plotter.plot(scene_event=scene_event, step_number=1)
        self.assertIsInstance(img, Image.Image)

    def test_plot_twice(self):
        metadata = {
            'screenWidth': 600,
            'screenHeight': 400,
            'objects': [],
            'structuralObjects': [],
            'agent': {
                'position': {
                    'x': 0.0,
                    'y': 0.0,
                    'z': 0.0
                },
                'rotation': {
                    'x': 0.0,
                    'y': 0.0,
                    'z': 0.0
                }
            }}
        scene_event = ai2thor.server.Event(metadata=metadata)
        img1 = self.plotter.plot(scene_event=scene_event, step_number=1)
        metadata = {
            'screenWidth': 600,
            'screenHeight': 400,
            'objects': [],
            'structuralObjects': [],
            'agent': {
                'position': {
                    'x': 2.0,
                    'y': 0.0,
                    'z': 2.0
                },
                'rotation': {
                    'x': 0.0,
                    'y': 0.0,
                    'z': 0.0
                }
            }}
        scene_event = ai2thor.server.Event(metadata=metadata)
        img2 = self.plotter.plot(scene_event=scene_event, step_number=2)
        self.assertNotEqual(img1, img2)
        self.assertIsInstance(img1, Image.Image)
        self.assertIsInstance(img2, Image.Image)

    def test_calculate_heading_zero_degrees(self):
        heading = self.plotter._calculate_heading(
            0, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.x, 0.0)
        self.assertAlmostEqual(heading.z, TopDownPlotter.HEADING_LENGTH)
        self.assertIsInstance(heading, XZHeading)

    def test_calculate_heading_ninety_degrees(self):
        heading = self.plotter._calculate_heading(
            90, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.x, -TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.z, 0.0)
        self.assertIsInstance(heading, XZHeading)

    def test_calculate_heading_one_eighty_degrees(self):
        heading = self.plotter._calculate_heading(
            180, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.x, 0.0)
        self.assertAlmostEqual(heading.z, -TopDownPlotter.HEADING_LENGTH)
        self.assertIsInstance(heading, XZHeading)

    def test_calculate_heading_two_seventy_degrees(self):
        heading = self.plotter._calculate_heading(
            270, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.x, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.z, 0.0)
        self.assertIsInstance(heading, XZHeading)

    def test_calculate_heading_three_sixty_degrees(self):
        heading = self.plotter._calculate_heading(
            360, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.x, 0.0)
        self.assertAlmostEqual(heading.z, TopDownPlotter.HEADING_LENGTH)
        self.assertIsInstance(heading, XZHeading)

    def test_calculate_heading_seven_twenty_degrees(self):
        heading = self.plotter._calculate_heading(
            720, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.x, 0.0)
        self.assertAlmostEqual(heading.z, TopDownPlotter.HEADING_LENGTH)
        self.assertIsInstance(heading, XZHeading)

    def test_calculate_heading_negative_ninety_degrees(self):
        heading = self.plotter._calculate_heading(
            -90, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.x, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.z, 0.0)
        self.assertIsInstance(heading, XZHeading)

    def test_calculate_heading_negative_seven_twenty_degrees(self):
        heading = self.plotter._calculate_heading(
            -720, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.x, 0.0)
        self.assertAlmostEqual(heading.z, TopDownPlotter.HEADING_LENGTH)
        self.assertIsInstance(heading, XZHeading)

    def test_create_robot(self):
        robot_metadata = {
            'position': {
                'x': 2.70,
                'y': 1.23,
                'z': 3.14
            },
            'rotation': {
                'x': 0.0,
                'y': 78.0,
                'z': 0.0
            }
        }
        robot = self.plotter._create_robot(robot_metadata)

        self.assertAlmostEqual(robot.x, 2.70)
        self.assertAlmostEqual(robot.z, 3.14)
        self.assertAlmostEqual(robot.rotation_y, 282)

    def test_create_robot_missing_position(self):
        robot_metadata = {
            'rotation': {
                'x': 0.0,
                'y': 78.0,
                'z': 0.0
            }
        }
        robot = self.plotter._create_robot(robot_metadata)

        self.assertAlmostEqual(robot.x, 0.0)
        self.assertAlmostEqual(robot.z, 0.0)
        self.assertAlmostEqual(robot.rotation_y, 282)

    def test_create_robot_missing_rotation(self):
        robot_metadata = {
            'position': {
                'x': 2.70,
                'y': 1.23,
                'z': 3.14
            }
        }
        robot = self.plotter._create_robot(robot_metadata)

        self.assertAlmostEqual(robot.x, 2.70)
        self.assertAlmostEqual(robot.z, 3.14)
        self.assertAlmostEqual(robot.rotation_y, 360)

    def test_create_object(self):
        object_metadata = {
            'isPickedUp': True,
            'visibleInCamera': True,
            'objectId': 'test-uuid',
            'colorsFromMaterials': [
                'orange'
            ],
            'objectBounds': {
                'objectBoundsCorners': [
                    {'x': 0, 'y': 0, 'z': 0},
                    {'x': 0, 'y': 0, 'z': 1},
                    {'x': 1, 'y': 0, 'z': 1},
                    {'x': 1, 'y': 0, 'z': 0},
                    {'x': 0, 'y': 1, 'z': 0},
                    {'x': 0, 'y': 1, 'z': 1},
                    {'x': 1, 'y': 1, 'z': 1},
                    {'x': 1, 'y': 1, 'z': 0}
                ]
            }
        }
        obj = self.plotter._create_asset(object_metadata)

        self.assertTrue(obj.held)
        self.assertTrue(obj.visible)
        self.assertEqual(obj.uuid, 'test-uuid')
        self.assertIsInstance(obj.bounds, SceneBounds)
        self.assertEqual(len(obj.bounds.points), 8)
        self.assertEqual(obj.color, "orange")

    def test_create_object_empty(self):
        object_metadata = {}
        obj = self.plotter._create_asset(object_metadata)

        self.assertIsNone(obj.held)
        self.assertIsNone(obj.visible)
        self.assertIsNone(obj.uuid)
        self.assertIsNone(obj.bounds)
        self.assertEqual(obj.color, "ivory")

    def test_find_plottable_objects_empty(self):
        metadata = {
            'screenWidth': 600,
            'screenHeight': 400,
            'objects': [],
            'structuralObjects': [],
            'agent': {
                'position': {
                    'x': 0.0,
                    'y': 0.0,
                    'z': 0.0
                },
                'rotation': {
                    'x': 0.0,
                    'y': 0.0,
                    'z': 0.0
                }
            }
        }
        scene_event = ai2thor.server.Event(metadata=metadata)
        filtered_objects = self.plotter._find_plottable_objects(scene_event)
        self.assertEqual(len(filtered_objects), 0)

    def test_find_plottable_objects_combined(self):
        metadata = {
            'screenWidth': 600,
            'screenHeight': 400,
            'objects': [
                {'objectId': 'test-uuid1',
                 'objectBounds': {'objectBoundsCorners': [{'y': 0}]}},
                {'objectId': 'test-uuid2',
                 'objectBounds': {'objectBoundsCorners': [{'y': 0}]}},
                {'objectId': 'test-uuid3',
                 'objectBounds': {'objectBoundsCorners': [{'y': 0}]}},
            ],
            'structuralObjects': [
                {'objectId': 'occluder1',
                 'objectBounds': {'objectBoundsCorners': [{'y': 0}]}},
                {'objectId': 'occluder2',
                 'objectBounds': {'objectBoundsCorners': [{'y': 0}]}},
                {'objectId': 'wall1',
                 'objectBounds': {'objectBoundsCorners': [{'y': 0}]}},
            ]
        }
        scene_event = ai2thor.server.Event(metadata=metadata)
        filtered_objects = self.plotter._find_plottable_objects(scene_event)
        self.assertEqual(len(filtered_objects), 6)
        self.assertEqual(len(
            [k for k in filtered_objects
             if k['objectId'].startswith('test-uuid')]), 3)
        self.assertEqual(len(
            [k for k in filtered_objects
             if k['objectId'].startswith('occluder')]), 2)
        self.assertEqual(len(
            [k for k in filtered_objects
             if k['objectId'].startswith('wall')]), 1)

    def test_find_plottable_objects_filtered(self):
        metadata = {
            'screenWidth': 600,
            'screenHeight': 400,
            'objects': [
                {'objectId': 'test-uuid1',
                 'objectBounds': {'objectBoundsCorners': [{'y': 0}]}},
                {'objectId': 'test-uuid2',
                 'objectBounds': {'objectBoundsCorners': [{'y': 0}]}},
                {'objectId': 'test-uuid3',
                 'objectBounds': {'objectBoundsCorners': [{'y': 0}]}},
            ],
            'structuralObjects': [
                {'objectId': 'occluder1',
                 'objectBounds': {'objectBoundsCorners': [{'y': 0}]}},
                {'objectId': 'occluder2',
                 'objectBounds': {'objectBoundsCorners': [{'y': 0}]}},
                {'objectId': 'wall1',
                 'objectBounds': {'objectBoundsCorners': [{'y': 0}]}},
                {'objectId': 'future-object',
                 'objectBounds': {'objectBoundsCorners': [{'y': 0}]}},
                {'objectId': 'floor-gets-filtered',
                 'objectBounds': {'objectBoundsCorners': [{'y': 0}]}},
                {'objectId': 'ceiling-gets-filtered',
                 'objectBounds': {'objectBoundsCorners': [{'y': 0}]}},
            ]
        }
        scene_event = ai2thor.server.Event(metadata=metadata)
        filtered_objects = self.plotter._find_plottable_objects(scene_event)
        self.assertEqual(len(filtered_objects), 7)
        self.assertEqual(filtered_objects[0]['objectId'], 'occluder1')
        self.assertEqual(filtered_objects[1]['objectId'], 'occluder2')
        self.assertEqual(filtered_objects[2]['objectId'], 'wall1')
        self.assertEqual(filtered_objects[3]['objectId'], 'future-object')
        self.assertEqual(filtered_objects[4]['objectId'], 'test-uuid1')
        self.assertEqual(filtered_objects[5]['objectId'], 'test-uuid2')
        self.assertEqual(filtered_objects[6]['objectId'], 'test-uuid3')

    def test_find_plottable_objects_sorted(self):
        metadata = {
            'screenWidth': 600,
            'screenHeight': 400,
            'objects': [
                {'objectId': 'test-uuid1',
                 'objectBounds': {'objectBoundsCorners': [{'y': 5}]}},
                {'objectId': 'test-uuid2',
                 'objectBounds': {'objectBoundsCorners': [{'y': 6}]}},
                {'objectId': 'test-uuid3',
                 'objectBounds': {'objectBoundsCorners': [{'y': 7}]}},
            ],
            'structuralObjects': [
                {'objectId': 'occluder1',
                 'objectBounds': {'objectBoundsCorners': [{'y': 1}]}},
                {'objectId': 'occluder2',
                 'objectBounds': {'objectBoundsCorners': [{'y': 2}]}},
                {'objectId': 'wall1',
                 'objectBounds': {'objectBoundsCorners': [{'y': 3}]}},
                {'objectId': 'future-object',
                 'objectBounds': {'objectBoundsCorners': [{'y': 4}]}},
                {'objectId': 'floor-gets-filtered',
                 'objectBounds': {'objectBoundsCorners': [{'y': 0}]}},
                {'objectId': 'ceiling-gets-filtered',
                 'objectBounds': {'objectBoundsCorners': [{'y': 0}]}},
            ]
        }
        scene_event = ai2thor.server.Event(metadata=metadata)
        filtered_objects = self.plotter._find_plottable_objects(scene_event)
        self.assertEqual(len(filtered_objects), 7)
        self.assertEqual(filtered_objects[0]['objectId'], 'occluder1')
        self.assertEqual(filtered_objects[1]['objectId'], 'occluder2')
        self.assertEqual(filtered_objects[2]['objectId'], 'wall1')
        self.assertEqual(filtered_objects[3]['objectId'], 'future-object')
        self.assertEqual(filtered_objects[4]['objectId'], 'test-uuid1')
        self.assertEqual(filtered_objects[5]['objectId'], 'test-uuid2')
        self.assertEqual(filtered_objects[6]['objectId'], 'test-uuid3')

    def test_scene_name(self):
        self.assertEqual(self.plotter._scene_name, "test")

    def test_scene_name_prefix(self):
        scene_config = SceneConfiguration(
            name="prefix/test", room_dimensions=Vector3d(
                x=10, y=4, z=10))
        plotter = TopDownPlotter(
            team="test",
            scene_config=scene_config
        )

        self.assertEqual(plotter._scene_name, "test")

    def test_draw_holes_even_room_dimensions(self):
        holes = [
            Vector2dInt(**{"x": 0, "z": 0}),
            Vector2dInt(**{"x": -3, "z": 0}),
            Vector2dInt(**{"x": 3, "z": 0}),
            Vector2dInt(**{"x": 0, "z": 4}),
            Vector2dInt(**{"x": 0, "z": -4}),
            Vector2dInt(**{"x": 2, "z": 0}),
            Vector2dInt(**{"x": -2, "z": 0}),
            Vector2dInt(**{"x": 2, "z": 3}),
            Vector2dInt(**{"x": -2, "z": -3}),
            Vector2dInt(**{"x": 2, "z": -3}),
            Vector2dInt(**{"x": -2, "z": 3}),
            Vector2dInt(**{"x": -3, "z": -4}),
            Vector2dInt(**{"x": 3, "z": 4}),
            Vector2dInt(**{"x": -3, "z": 4}),
            Vector2dInt(**{"x": 3, "z": -4}),
        ]
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={'metadata': {}},
            holes=holes,
            room_dimensions=Vector3d(
                x=6,
                y=3,
                z=8),
            floor_textures=[]
        )

        plotter = TopDownPlotter(
            team="test",
            scene_config=scene_config)
        img = plotter.base_room_img.copy()
        holes_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        """
        holes_img.save(
            os.path.join(
                resources_path,
                'plotter_holes_even_dimensions.png'))
        """

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(
                resources_path,
                'plotter_holes_even_dimensions.png'))
        # calculate image difference
        diff = ImageChops.difference(holes_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_draw_holes_odd_room_dimensions(self):
        holes = [
            Vector2dInt(**{"x": 0, "z": 0}),
            Vector2dInt(**{"x": 2, "z": 3}),
            Vector2dInt(**{"x": 2, "z": -3}),
            Vector2dInt(**{"x": -2, "z": -3}),
            Vector2dInt(**{"x": -2, "z": 3}),
            Vector2dInt(**{"x": 2, "z": 1}),
            Vector2dInt(**{"x": -2, "z": 1}),
            Vector2dInt(**{"x": -2, "z": -1}),
            Vector2dInt(**{"x": 2, "z": -1}),
            Vector2dInt(**{"x": 0, "z": 3}),
            Vector2dInt(**{"x": 0, "z": -3}),
            Vector2dInt(
                **{"x": -3, "z": 3}),  # out of bounds - don't draw
            Vector2dInt(
                **{"x": 3, "z": -3}),  # out of bounds - don't draw
        ]
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={'metadata': {}},
            holes=holes,
            room_dimensions=Vector3d(
                x=5,
                y=3,
                z=7),
            floor_textures=[]
        )

        plotter = TopDownPlotter(
            team="test",
            scene_config=scene_config)
        img = plotter.base_room_img.copy()
        holes_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        """
        holes_img.save(
            os.path.join(
                resources_path,
                'plotter_holes_odd_dimensions.png'))
        """

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(
                resources_path,
                'plotter_holes_odd_dimensions.png'))
        # calculate image difference
        diff = ImageChops.difference(holes_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_draw_holes_even_odd_room_dimensions(self):
        holes = [
            Vector2dInt(**{"x": 0, "z": 0}),
            Vector2dInt(**{"x": 5, "z": 4}),
            Vector2dInt(**{"x": -5, "z": 4}),
            Vector2dInt(**{"x": 5, "z": -4}),
            Vector2dInt(**{"x": -5, "z": -4}),
            Vector2dInt(**{"x": 4, "z": -4}),
            Vector2dInt(**{"x": 4, "z": -3}),
            Vector2dInt(**{"x": 5, "z": -3}),
            Vector2dInt(**{"x": -5, "z": -3}),
            Vector2dInt(**{"x": -4, "z": -3}),
            Vector2dInt(**{"x": -4, "z": -4}),
            Vector2dInt(**{"x": 0, "z": -4}),
            Vector2dInt(**{"x": 0, "z": 4}),
            Vector2dInt(**{"x": 4, "z": 0}),
            Vector2dInt(**{"x": -4, "z": 0}),
            Vector2dInt(**{"x": 5, "z": 3}),
            Vector2dInt(**{"x": -5, "z": 3}),
            Vector2dInt(**{"x": -4, "z": 3}),
            Vector2dInt(**{"x": 4, "z": 3}),
            Vector2dInt(**{"x": 4, "z": 4}),
            Vector2dInt(**{"x": -4, "z": 4}),
            Vector2dInt(**{"x": -5, "z": 0}),
            Vector2dInt(**{"x": 5, "z": 0}),
            Vector2dInt(
                **{"x": 6, "z": 0}),  # out of bounds - dont draw
            Vector2dInt(
                **{"x": -5, "z": -5})  # out of bounds - dont draw

        ]
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={'metadata': {}},
            holes=holes,
            room_dimensions=Vector3d(
                x=10,
                y=3,
                z=9),
            lava=[]
        )

        plotter = TopDownPlotter(
            team="test",
            scene_config=scene_config)
        img = plotter.base_room_img.copy()
        holes_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        """
        holes_img.save(
            os.path.join(
                resources_path,
                'plotter_holes_even_odd_dimensions.png'))
        """

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(
                resources_path,
                'plotter_holes_even_odd_dimensions.png'))
        # calculate image difference
        diff = ImageChops.difference(holes_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_draw_lava(self):
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={'metadata': {}},
            room_dimensions=Vector3d(
                x=6,
                y=3,
                z=10),
            lava=[
                Vector2dInt(**{"x": 0, "z": 0}),
                Vector2dInt(**{"x": -3, "z": 0}),
                Vector2dInt(**{"x": 3, "z": 0}),
                Vector2dInt(**{"x": 0, "z": 5}),
                Vector2dInt(**{"x": 0, "z": -5}),
                Vector2dInt(**{"x": 2, "z": 0}),
                Vector2dInt(**{"x": -2, "z": 0}),
                Vector2dInt(**{"x": 2, "z": 4}),
                Vector2dInt(**{"x": -2, "z": -4}),
                Vector2dInt(**{"x": 2, "z": -4}),
                Vector2dInt(**{"x": -2, "z": 4}),
                Vector2dInt(**{"x": -3, "z": -4}),
                Vector2dInt(**{"x": 3, "z": 4}),
                Vector2dInt(**{"x": -3, "z": 4}),
                Vector2dInt(**{"x": 3, "z": -4}),
                Vector2dInt(**{"x": 3, "z": -5}),
                Vector2dInt(**{"x": 3, "z": 5}),
                Vector2dInt(**{"x": -3, "z": 5}),
                Vector2dInt(**{"x": -3, "z": -5}),
                Vector2dInt(**{"x": 2, "z": -5}),
                Vector2dInt(**{"x": 2, "z": 5}),
                Vector2dInt(**{"x": -2, "z": 5}),
                Vector2dInt(**{"x": -2, "z": -5}),
                Vector2dInt(
                    **{"x": -4, "z": -6}),  # out of bounds - lava don't draw
                Vector2dInt(
                    **{"x": 4, "z": 6})  # out of bounds bad - lava don't draw
            ]
        )

        plotter = TopDownPlotter(
            team="test",
            scene_config=scene_config)
        img = plotter.base_room_img.copy()

        lava_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        # lava_img.save(os.path.join(resources_path, 'plotter_lava.png'))

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(
                resources_path,
                'plotter_lava.png'))
        # calculate image difference
        diff = ImageChops.difference(lava_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_draw_lava_from_partition_floor(self):
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={'metadata': {}},
            room_dimensions=Vector3d(x=10, y=3, z=10),
            partition_floor=FloorPartitionConfig(left_half=0.2, right_half=0.8)
        )

        plotter = TopDownPlotter(team="test", scene_config=scene_config)
        img = plotter.base_room_img.copy()

        lava_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        # lava_img.save(os.path.join(resources_path, 'plotter_lava_2.png'))

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(resources_path, 'plotter_lava_2.png')
        )
        # calculate image difference
        diff = ImageChops.difference(lava_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_draw_room(self):
        '''Floor grid plus walls'''
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={'metadata': {}},
            room_dimensions=Vector3d(
                x=10,
                y=3,
                z=10),
            lava=[]
        )

        plotter = TopDownPlotter(
            team="test",
            scene_config=scene_config)
        img = plotter.base_room_img.copy()
        base_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        # base_img.save(os.path.join(resources_path, 'plotter_base.png'))

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(
                resources_path,
                'plotter_base.png'))
        # calculate image difference
        diff = ImageChops.difference(base_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_draw_nonsquare_room(self):
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={'metadata': {}},
            room_dimensions=Vector3d(
                x=10,
                y=3,
                z=20),
            lava=[]
        )

        plotter = TopDownPlotter(
            team="test",
            scene_config=scene_config)
        img = plotter.base_room_img.copy()
        base_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        # base_img.save(os.path.join(
        #     resources_path,
        #     'plotter_nonsquare_base.png'))

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(
                resources_path,
                'plotter_nonsquare_base.png'))
        # calculate image difference
        diff = ImageChops.difference(base_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_draw_agent_facing_angled(self):
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={},
            room_dimensions=Vector3d(x=10, y=3, z=10),
            lava=[]
        )
        robot_metadata = {
            'position': {'x': 1, 'y': 0, 'z': 2},
            'rotation': {'x': 0, 'y': 135, 'z': 0}
        }

        plotter = TopDownPlotter(team="test", scene_config=scene_config)
        img = plotter.base_room_img.copy()
        img = plotter._draw_robot(img, robot_metadata=robot_metadata)
        agent_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        # agent_img.save(
        #     os.path.join(resources_path, 'plotter_agent_angle.png')
        # )

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(resources_path, 'plotter_agent_angle.png')
        )
        # calculate image difference
        diff = ImageChops.difference(agent_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_draw_agent_facing_front(self):
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={},
            room_dimensions=Vector3d(x=10, y=3, z=10),
            lava=[]
        )
        robot_metadata = {
            'position': {'x': 1, 'y': 0, 'z': 2},
            'rotation': {'x': 0, 'y': 0, 'z': 0}
        }

        plotter = TopDownPlotter(team="test", scene_config=scene_config)
        img = plotter.base_room_img.copy()
        img = plotter._draw_robot(img, robot_metadata=robot_metadata)
        agent_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        # agent_img.save(
        #     os.path.join(resources_path, 'plotter_agent_front.png')
        # )

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(resources_path, 'plotter_agent_front.png')
        )
        # calculate image difference
        diff = ImageChops.difference(agent_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_draw_agent_facing_left(self):
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={},
            room_dimensions=Vector3d(x=10, y=3, z=10),
            lava=[]
        )
        robot_metadata = {
            'position': {'x': 1, 'y': 0, 'z': 2},
            'rotation': {'x': 0, 'y': -90, 'z': 0}
        }

        plotter = TopDownPlotter(team="test", scene_config=scene_config)
        img = plotter.base_room_img.copy()
        img = plotter._draw_robot(img, robot_metadata=robot_metadata)
        agent_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        # agent_img.save(
        #     os.path.join(resources_path, 'plotter_agent_left.png')
        # )

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(resources_path, 'plotter_agent_left.png')
        )
        # calculate image difference
        diff = ImageChops.difference(agent_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_draw_goal_object(self):
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={'metadata': {}},
            room_dimensions=Vector3d(
                x=10,
                y=3,
                z=10),
            lava=[]
        )
        obj = SceneAsset(
            held=False,
            visible=True,
            uuid='1',
            color='green',
            bounds=SceneBounds([SceneCoord(**pt) for pt in [
                {'x': 1, 'y': 0, 'z': 1},
                {'x': 2, 'y': 0, 'z': 1},
                {'x': 2, 'y': 0, 'z': 2},
                {'x': 1, 'y': 0, 'z': 2},
                {'x': 1, 'y': 0.22, 'z': 1},
                {'x': 2, 'y': 0.22, 'z': 1},
                {'x': 2, 'y': 0.22, 'z': 2},
                {'x': 1, 'y': 0.22, 'z': 2}
            ]]))

        plotter = TopDownPlotter(
            team="test",
            scene_config=scene_config)
        img = plotter.base_room_img.copy()
        img = plotter._draw_goal(img, obj)
        goal_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        # goal_img.save(os.path.join(resources_path, 'plotter_goal.png'))

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(
                resources_path,
                'plotter_goal.png'))
        # calculate image difference
        diff = ImageChops.difference(goal_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_draw_goal_object_greyscale(self):
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={},
            room_dimensions=Vector3d(x=10, y=3, z=10),
            lava=[]
        )
        obj = SceneAsset(
            held=False,
            visible=True,
            uuid='1',
            color='white',
            bounds=SceneBounds([SceneCoord(**pt) for pt in [
                {'x': -1, 'y': 0, 'z': -1},
                {'x': -1.5, 'y': 0, 'z': -1},
                {'x': -1.5, 'y': 0, 'z': -1.5},
                {'x': -1, 'y': 0, 'z': -1.5},
                {'x': -1, 'y': 0.22, 'z': -1},
                {'x': -1.5, 'y': 0.22, 'z': -1},
                {'x': -1.5, 'y': 0.22, 'z': -1.5},
                {'x': -1, 'y': 0.22, 'z': -1.5}
            ]]))

        plotter = TopDownPlotter(team="test", scene_config=scene_config)
        img = plotter.base_room_img.copy()
        img = plotter._draw_goal(img, obj)
        goal_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        # goal_img.save(os.path.join(resources_path, 'plotter_goal_white.png'))

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(resources_path, 'plotter_goal_white.png')
        )
        # calculate image difference
        diff = ImageChops.difference(goal_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_draw_goal_object_angled(self):
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={},
            room_dimensions=Vector3d(x=10, y=3, z=10),
            lava=[]
        )
        obj = SceneAsset(
            held=False,
            visible=True,
            uuid='1',
            color='blue',
            bounds=SceneBounds([SceneCoord(**pt) for pt in [
                {'x': 0.5, 'y': 0, 'z': 0},
                {'x': 0, 'y': 0, 'z': -0.5},
                {'x': -0.5, 'y': 0, 'z': 0},
                {'x': 0, 'y': 0, 'z': 0.5},
                {'x': 0.5, 'y': 0.22, 'z': 0},
                {'x': 0, 'y': 0.22, 'z': -0.5},
                {'x': -0.5, 'y': 0.22, 'z': 0},
                {'x': 0, 'y': 0.22, 'z': 0.5}
            ]]))

        plotter = TopDownPlotter(team="test", scene_config=scene_config)
        img = plotter.base_room_img.copy()
        img = plotter._draw_goal(img, obj)
        goal_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        # goal_img.save(os.path.join(resources_path, 'plotter_goal_angle.png'))

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(resources_path, 'plotter_goal_angle.png')
        )
        # calculate image difference
        diff = ImageChops.difference(goal_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_draw_goal_object_hidden(self):
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={},
            room_dimensions=Vector3d(x=10, y=3, z=10),
            lava=[]
        )
        obj = SceneAsset(
            held=False,
            visible=False,
            uuid='1',
            color='blue',
            bounds=SceneBounds([SceneCoord(**pt) for pt in [
                {'x': 0.5, 'y': 0, 'z': 0},
                {'x': 0, 'y': 0, 'z': -0.5},
                {'x': -0.5, 'y': 0, 'z': 0},
                {'x': 0, 'y': 0, 'z': 0.5},
                {'x': 0.5, 'y': 0.22, 'z': 0},
                {'x': 0, 'y': 0.22, 'z': -0.5},
                {'x': -0.5, 'y': 0.22, 'z': 0},
                {'x': 0, 'y': 0.22, 'z': 0.5}
            ]]))

        plotter = TopDownPlotter(team="test", scene_config=scene_config)
        img = plotter.base_room_img.copy()
        img = plotter._draw_goal(img, obj)
        goal_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        # goal_img.save(
        #     os.path.join(resources_path, 'plotter_goal_hidden.png')
        # )

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(resources_path, 'plotter_goal_hidden.png')
        )
        # calculate image difference
        diff = ImageChops.difference(goal_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_draw_visible_object(self):
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={'metadata': {}},
            room_dimensions=Vector3d(
                x=10,
                y=3,
                z=10),
            lava=[]
        )
        obj = SceneAsset(
            held=False,
            visible=True,
            uuid='1',
            color='blue',
            bounds=SceneBounds([SceneCoord(**pt) for pt in [
                {'x': 1, 'y': 0, 'z': 1},
                {'x': 2, 'y': 0, 'z': 1},
                {'x': 2, 'y': 0, 'z': 2},
                {'x': 1, 'y': 0, 'z': 2},
                {'x': 1, 'y': 0.22, 'z': 1},
                {'x': 2, 'y': 0.22, 'z': 1},
                {'x': 2, 'y': 0.22, 'z': 2},
                {'x': 1, 'y': 0.22, 'z': 2}
            ]]))

        plotter = TopDownPlotter(
            team="test",
            scene_config=scene_config)
        img = plotter.base_room_img.copy()
        img = plotter._draw_object(img, obj)
        visible_object_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        # visible_object_img.save(os.path.join(
        #     resources_path,
        #     'plotter_visible_object.png'))

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(
                resources_path,
                'plotter_visible_object.png'))
        # calculate image difference
        diff = ImageChops.difference(visible_object_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_draw_visible_object_angled(self):
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={},
            room_dimensions=Vector3d(x=10, y=3, z=10),
            lava=[]
        )
        obj = SceneAsset(
            held=False,
            visible=True,
            uuid='1',
            color='brown',
            bounds=SceneBounds([SceneCoord(**pt) for pt in [
                {'x': 2, 'y': 0, 'z': 1},
                {'x': -1, 'y': 0, 'z': -2},
                {'x': -2, 'y': 0, 'z': -1},
                {'x': 1, 'y': 0, 'z': 2},
                {'x': 2, 'y': 0.22, 'z': 1},
                {'x': -1, 'y': 0.22, 'z': -2},
                {'x': -2, 'y': 0.22, 'z': -1},
                {'x': 1, 'y': 0.22, 'z': 2}
            ]]))

        plotter = TopDownPlotter(team="test", scene_config=scene_config)
        img = plotter.base_room_img.copy()
        img = plotter._draw_object(img, obj)
        visible_object_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        # visible_object_img.save(
        #     os.path.join(resources_path, 'plotter_visible_object_angle.png')
        # )

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(resources_path, 'plotter_visible_object_angle.png')
        )
        # calculate image difference
        diff = ImageChops.difference(visible_object_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_draw_hidden_object(self):
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={'metadata': {}},
            room_dimensions=Vector3d(
                x=10,
                y=3,
                z=10),
            lava=[]
        )
        obj = SceneAsset(
            held=False,
            visible=False,
            uuid='1',
            color='yellow',
            bounds=SceneBounds([SceneCoord(**pt) for pt in [
                {'x': 1, 'y': 0, 'z': 1},
                {'x': 2, 'y': 0, 'z': 1},
                {'x': 2, 'y': 0, 'z': 2},
                {'x': 1, 'y': 0, 'z': 2},
                {'x': 1, 'y': 0.22, 'z': 1},
                {'x': 2, 'y': 0.22, 'z': 1},
                {'x': 2, 'y': 0.22, 'z': 2},
                {'x': 1, 'y': 0.22, 'z': 2}
            ]]))

        plotter = TopDownPlotter(
            team="test",
            scene_config=scene_config)
        img = plotter.base_room_img.copy()
        img = plotter._draw_object(img, obj)
        hidden_object_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        # hidden_object_img.save(os.path.join(
        #     resources_path,
        #     'plotter_hidden_object.png'))

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(
                resources_path,
                'plotter_hidden_object.png'))
        # calculate image difference
        diff = ImageChops.difference(hidden_object_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_draw_hidden_object_angled(self):
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={},
            room_dimensions=Vector3d(x=10, y=3, z=10),
            lava=[]
        )
        obj = SceneAsset(
            held=False,
            visible=False,
            uuid='1',
            color='brown',
            bounds=SceneBounds([SceneCoord(**pt) for pt in [
                {'x': 2, 'y': 0, 'z': 1},
                {'x': -1, 'y': 0, 'z': -2},
                {'x': -2, 'y': 0, 'z': -1},
                {'x': 1, 'y': 0, 'z': 2},
                {'x': 2, 'y': 0.22, 'z': 1},
                {'x': -1, 'y': 0.22, 'z': -2},
                {'x': -2, 'y': 0.22, 'z': -1},
                {'x': 1, 'y': 0.22, 'z': 2}
            ]]))

        plotter = TopDownPlotter(team="test", scene_config=scene_config)
        img = plotter.base_room_img.copy()
        img = plotter._draw_object(img, obj)
        hidden_object_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        # hidden_object_img.save(
        #     os.path.join(resources_path, 'plotter_hidden_object_angle.png')
        # )

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(resources_path, 'plotter_hidden_object_angle.png')
        )
        # calculate image difference
        diff = ImageChops.difference(hidden_object_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_draw_ramp_arrow(self):
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={'metadata': {}},
            room_dimensions=Vector3d(
                x=10,
                y=3,
                z=10),
            lava=[]
        )
        bounds = SceneBounds([SceneCoord(**pt) for pt in [
            {'x': 1.5, 'y': 0, 'z': 2},
            {'x': 2.5, 'y': 0, 'z': 2},
            {'x': 2.5, 'y': 0, 'z': -2},
            {'x': 1.5, 'y': 0, 'z': -2},
            {'x': 1.5, 'y': 2, 'z': 2},
            {'x': 2.5, 'y': 2, 'z': 2},
            {'x': 2.5, 'y': 2, 'z': -2},
            {'x': 1.5, 'y': 2, 'z': -2}
        ]])

        ramp = SceneAsset(
            held=False,
            visible=True,
            uuid='ramp-1',
            color='yellow',
            bounds=bounds
        )

        plotter = TopDownPlotter(
            team="test",
            scene_config=scene_config)
        img = plotter.base_room_img.copy()
        img = plotter._draw_object(img, ramp)
        ramp_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        # ramp_img.save(os.path.join(
        #     resources_path,
        #     'plotter_ramp.png'))

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(
                resources_path,
                'plotter_ramp.png'))
        # calculate image difference
        diff = ImageChops.difference(ramp_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_draw_hidden_ramp_arrow(self):
        scene_config = SceneConfiguration(
            name="testscene",
            version=1,
            goal={'metadata': {}},
            room_dimensions=Vector3d(
                x=10,
                y=3,
                z=10),
            lava=[]
        )
        bounds = SceneBounds([SceneCoord(**pt) for pt in [
            {'x': 1.5, 'y': 0, 'z': 2},
            {'x': 2.5, 'y': 0, 'z': 2},
            {'x': 2.5, 'y': 0, 'z': -2},
            {'x': 1.5, 'y': 0, 'z': -2},
            {'x': 1.5, 'y': 2, 'z': 2},
            {'x': 2.5, 'y': 2, 'z': 2},
            {'x': 2.5, 'y': 2, 'z': -2},
            {'x': 1.5, 'y': 2, 'z': -2}
        ]])

        ramp = SceneAsset(
            held=False,
            visible=False,
            uuid='ramp-1',
            color='yellow',
            bounds=bounds
        )

        plotter = TopDownPlotter(
            team="test",
            scene_config=scene_config)
        img = plotter.base_room_img.copy()
        img = plotter._draw_object(img, ramp)
        ramp_img = plotter._export_plot(img)
        # save image to resources folder in the event of plotter changes
        # ramp_img.save(os.path.join(
        #     resources_path,
        #     'plotter_ramp_hidden.png'))

        # read image from resources folder
        truth_img = Image.open(
            os.path.join(
                resources_path,
                'plotter_ramp_hidden.png'))
        # calculate image difference
        diff = ImageChops.difference(ramp_img, truth_img)
        stat = ImageStat.Stat(diff)
        self.assertListEqual(stat.sum, [0.0, 0.0, 0.0])

    def test_find_opposite_color(self):
        color = self.plotter._find_opposite_color((0, 128, 255))
        self.assertTupleEqual(color, (255, 127, 0))


if __name__ == '__main__':
    unittest.main()



import random
import string
from typing import Any, Dict, List, Optional, Tuple

from .action import (FORCE_ACTIONS, OBJECT_IMAGE_ACTIONS, OBJECT_MOVE_ACTIONS,
                     RECEPTACLE_ACTIONS, Action)
from .config_manager import ConfigManager, MetadataTier
from .controller import DEFAULT_MOVE


def compare_param_values(value_1: Any, value_2: Any) -> bool:
    """Compares two parameter values and returns if they are equal,
    making sure that string numbers are converted to floats, and integer
    floats are converted to ints."""
    data = {'value_1': value_1, 'value_2': value_2}
    for key, value in data.items():
        if isinstance(value, str):
            try:
                data[key] = float(value)
            except ValueError:
                ...
        if isinstance(value, float) and value.is_integer():
            data[key] = int(value)
    return data['value_1'] == data['value_2']


def rebuild_endhabituation(step_action_list: List) -> str:
    '''Rebuilds EndHabituation parameters from the goal's action list for the
    current step. Parameters can include some or all of xPosition, zPosition,
    and yRotation. Parameters are removed from the step_metadata list of
    potentital actions in order for teleportation/displacement to be hidden
    from AIs.
    '''
    # sourcery skip: use-named-expression
    action = Action.END_HABITUATION.value
    endhabituation_action = next((
        item for item in step_action_list
        if item[0] == action), None)
    if endhabituation_action is not None:
        params = ",".join(
            f"{k}={v}" for k,
            v in endhabituation_action[1].items())
        if params:
            action = f"{action},{params}"
    return action


class Parameter:

    # AI2-THOR creates a square grid across the scene that is
    # uses for "snap-to-grid" movement. (This value may not
    # really matter because we set continuous to True in
    # the step input.)
    GRID_SIZE = 0.1

    DEFAULT_HORIZON = 0.0
    DEFAULT_ROTATION = 0.0
    DEFAULT_AMOUNT = 0.5
    DEFAULT_IMG_COORD = None
    DEFAULT_IMG_DICT = {'x': 0, 'y': 0}
    DEFAULT_OBJECT_MOVE_AMOUNT = 1.0
    DEFAULT_OBJECT_ROTATION_CLOCKWISE = True
    DEFAULT_OBJECT_MOVEMENT_X_DIRECTION = 0
    DEFAULT_OBJECT_MOVEMENT_Z_DIRECTION = 1

    MAX_AMOUNT_TORQUE = 1.0
    MIN_AMOUNT_TORQUE = -1.0
    MAX_AMOUNT = 1.0
    MIN_AMOUNT = 0.0
    MIN_AMOUNT_MOVEMENT_DIRECTION = -1
    MAX_AMOUNT_MOVEMENT_DIRECTION = 1

    MIN_NOISE = -0.5
    MAX_NOISE = 0.5

    ROTATION_KEY = 'rotation'
    HORIZON_KEY = 'horizon'
    FORCE_KEY = 'force'
    AMOUNT_KEY = 'amount'
    CLOCKWISE_KEY = 'clockwise'
    MOVEMENT_X_DIRECTION_KEY = 'lateral'
    MOVEMENT_Z_DIRECTION_KEY = 'straight'

    OBJECT_IMAGE_COORDS_X_KEY = 'objectImageCoordsX'
    OBJECT_IMAGE_COORDS_Y_KEY = 'objectImageCoordsY'
    RECEPTACLE_IMAGE_COORDS_X_KEY = 'receptacleObjectImageCoordsX'
    RECEPTACLE_IMAGE_COORDS_Y_KEY = 'receptacleObjectImageCoordsY'

    # used for EndHabituation teleport
    TELEPORT_X_POS = 'xPosition'
    TELEPORT_Z_POS = 'zPosition'
    TELEPORT_Y_ROT = 'yRotation'

    def __init__(self, config: ConfigManager):
        self.config = config

    def wrap_step(self, output_folder, **kwargs) -> Dict:
        # whether or not to randomize segmentation mask colors
        metadata_tier = self.config.get_metadata_tier()
        consistent_colors = (metadata_tier == MetadataTier.ORACLE)
        # Create the step data dict for the AI2-THOR step function.
        return dict(
            continuous=True,
            gridSize=self.GRID_SIZE,
            logs=True,
            renderDepthImage=self.config.is_depth_maps_enabled(),
            renderObjectImage=self.config.is_object_masks_enabled(),
            snapToGrid=False,
            consistentColors=consistent_colors,
            recordTopDown=(self.config.is_video_enabled() and
                           self.config.is_top_down_camera()),
            topDownImagePath=output_folder,
            **kwargs
        )

    def build_ai2thor_step(self, output_path, **kwargs) -> Tuple:
        action, params = self._validate_and_convert_params(**kwargs)
        action = self._mcs_action_to_ai2thor_action(action)
        wrapped_step = self.wrap_step(
            output_folder=output_path,
            action=action,
            **params)
        return wrapped_step, params

    def _convert_y_image_coord_for_unity(self, y_coord: int) -> int:
        '''Pixel coordinates are expected to start at the top left,
        but in Unity, (0,0) is the bottom left.
        '''
        if not isinstance(y_coord, int):
            raise TypeError(f"{y_coord} y_coord is not an integer")
        screen_height = self.config.get_screen_height()
        if(0 <= y_coord < screen_height):
            return screen_height - y_coord - 1
        raise ValueError(f"{y_coord} is not in range 0-{screen_height-1}")

    def _get_amount(self, action: Action, **kwargs) -> float:
        amount = kwargs.get(
            self.AMOUNT_KEY,
            self.DEFAULT_OBJECT_MOVE_AMOUNT
            if action in OBJECT_MOVE_ACTIONS
            else self.DEFAULT_AMOUNT
        )
        if amount is not None:
            try:
                amount = float(amount)
            except ValueError as err:
                raise ValueError(f"Amount {amount} is not a number") from err
        else:
            amount = self.DEFAULT_AMOUNT

        if amount < self.MIN_AMOUNT or amount > self.MAX_AMOUNT:
            raise ValueError(
                "Amount not in acceptable range of "
                f"({self.MIN_AMOUNT}-{self.MAX_AMOUNT})"
            )
        return amount

    def _get_force(self, action: Action, **kwargs) -> float:
        force = kwargs.get(self.FORCE_KEY, self.DEFAULT_AMOUNT)
        if force is not None:
            try:
                force = float(force)
            except ValueError as err:
                raise ValueError('Force is not a number') from err

            if action == Action.TORQUE_OBJECT:
                if (force < self.MIN_AMOUNT_TORQUE or force >
                        self.MAX_AMOUNT_TORQUE):
                    raise ValueError(
                        f'Force not in acceptable range of '
                        f'({self.MIN_AMOUNT_TORQUE} and '
                        f'{self.MAX_AMOUNT_TORQUE})')

            elif force < self.MIN_AMOUNT or force > self.MAX_AMOUNT:
                raise ValueError(
                    f'Force not in acceptable range of '
                    f'({self.MIN_AMOUNT} and {self.MAX_AMOUNT})')
        else:
            force = self.DEFAULT_AMOUNT
        return force

    def _get_number(self, key: str, **kwargs) -> Optional[Any]:
        val = kwargs.get(key)
        if val is not None:
            try:
                val = float(val)
            except ValueError as err:
                raise ValueError(f"{key}") from err
        return val

    def _get_number_with_default(
            self, key: str, default: Any, **kwargs) -> Any:
        val = kwargs.get(key, default)
        if val is not None:
            try:
                val = float(val)
            except ValueError as err:
                raise ValueError(f"{key}") from err
        elif default is not None:
            val = float(default)
        return val

    def _get_clockwise(self, **kwargs) -> bool:
        direction_clockwise = kwargs.get(
            self.CLOCKWISE_KEY,
            self.DEFAULT_OBJECT_ROTATION_CLOCKWISE
        )
        if isinstance(direction_clockwise, str):
            direction_clockwise = direction_clockwise.capitalize()
            try:
                direction_clockwise = eval(direction_clockwise)
            except Exception as err:
                raise ValueError(
                    f"{direction_clockwise} is not a bool") from err
        elif not isinstance(direction_clockwise, bool):
            raise ValueError(
                f"{direction_clockwise} is not a bool")
        return direction_clockwise

    def _get_movement_direction(self, **kwargs) -> Tuple:
        """
        If no args are given, the default movement for (x,z) is (0,1)
        If only x arg is given, movement is (x,0)
        If only z arg is given, movement is (0,z)
        If x,z args are given, movement is (x,z)
        """

        lateral = kwargs.get(
            self.MOVEMENT_X_DIRECTION_KEY
        )
        straight = kwargs.get(
            self.MOVEMENT_Z_DIRECTION_KEY
        )
        l_none = lateral is None
        s_none = straight is None

        if l_none and s_none:
            lateral = self.DEFAULT_OBJECT_MOVEMENT_X_DIRECTION
            straight = self.DEFAULT_OBJECT_MOVEMENT_Z_DIRECTION
            return (lateral, straight)

        direction_output = (
            f"{'(lateral: ' f'{lateral})' if not l_none else ''}"
            f"{'' if l_none else ' and ' if not s_none else ' is'}"
            f"{'(straight: ' f'{straight})' if not s_none else ''}"
            f"{' is' if l_none else ' are' if not s_none else ''}"
            f" not "
            f"{'both ints' if not l_none and not s_none else 'an int'}"
            f" of acceptable range "
            f"({self.MIN_AMOUNT_MOVEMENT_DIRECTION}"
            f" and {self.MAX_AMOUNT_MOVEMENT_DIRECTION})")

        if isinstance(lateral, bool) or isinstance(straight, bool):
            raise ValueError(direction_output)

        try:
            if not l_none:
                lateral = int(lateral) if float(
                    lateral).is_integer() else lateral
                if isinstance(lateral, float):
                    raise ValueError(direction_output)
            if not s_none:
                straight = int(straight) if float(
                    straight).is_integer() else straight
                if isinstance(straight, float):
                    raise ValueError(direction_output)
        except Exception as err:
            raise ValueError(direction_output) from err

        if not l_none and s_none:
            if (lateral < self.MIN_AMOUNT_MOVEMENT_DIRECTION or
                    lateral > self.MAX_AMOUNT_MOVEMENT_DIRECTION):
                raise ValueError(direction_output)
            else:
                straight = 0
        elif l_none:
            if (straight < self.MIN_AMOUNT_MOVEMENT_DIRECTION or
                    straight > self.MAX_AMOUNT_MOVEMENT_DIRECTION):
                raise ValueError(direction_output)
            else:
                lateral = 0
        elif (lateral < self.MIN_AMOUNT_MOVEMENT_DIRECTION or
              lateral > self.MAX_AMOUNT_MOVEMENT_DIRECTION or
              straight < self.MIN_AMOUNT_MOVEMENT_DIRECTION or
              straight > self.MAX_AMOUNT_MOVEMENT_DIRECTION):
            raise ValueError(direction_output)
        return (lateral, straight)

    def _get_move_magnitude(self, action: Action, force: float,
                            amount: float) -> float:
        # Set the Move Magnitude to the appropriate amount based on the action
        move_magnitude = DEFAULT_MOVE
        if action in FORCE_ACTIONS:
            move_magnitude = force * self.MAX_AMOUNT
        elif action in OBJECT_MOVE_ACTIONS:
            move_magnitude = amount
        return move_magnitude

    def _get_teleport(self, **kwargs) -> Tuple:
        teleport_rotation = self._get_teleport_rotation(**kwargs)
        teleport_position = self._get_teleport_position(**kwargs)
        return (teleport_rotation, teleport_position)

    def _get_teleport_position(self, **kwargs) -> Optional[Dict]:
        '''Extract teleport xz position from kwargs if it exists.
        Otherwise, position will be None.
        '''
        teleport_pos_x_input = self._get_number(self.TELEPORT_X_POS, **kwargs)
        teleport_pos_z_input = self._get_number(self.TELEPORT_Z_POS, **kwargs)
        return {
            'x': teleport_pos_x_input,
            'z': teleport_pos_z_input} \
            if teleport_pos_x_input is not None and \
            teleport_pos_z_input is not None else None

    def _get_teleport_rotation(self, **kwargs) -> Optional[Dict]:
        '''Extract teleport rotation from kwargs if it exists.
        Otherwise, rotation will be None.
        '''
        teleport_rot_input = self._get_number(self.TELEPORT_Y_ROT, **kwargs)
        return {'y': teleport_rot_input} \
            if teleport_rot_input is not None else None

    def _get_vector(
            self,
            x_key: string,
            y_key: string,
            default_coord: int,
            object_id: string,
            **kwargs) -> Dict:
        image_coords_x = self._get_number_with_default(
            x_key, default_coord, **kwargs)
        image_coords_y = self._get_number_with_default(
            y_key, default_coord, **kwargs)

        # If the user passes no x or y parameter and no object Id,
        #   throw an error, else if an object id only is sent, pass
        #   the default values of 0,0 so Unity doesn't blow up.
        if (image_coords_x is None and image_coords_y is None and
                kwargs.get(object_id) is None):
            raise Exception('MCS Action Failed to provide coordinate value')
        elif (image_coords_x is None and image_coords_y is None and
                kwargs.get(object_id) is not None):
            return self.DEFAULT_IMG_DICT

        return {
            'x': int(image_coords_x),
            'y': self._convert_y_image_coord_for_unity(
                int(image_coords_y))
        }

    def _get_receptacle_vector(self, **kwargs) -> Dict:
        return self._get_vector(
            self.RECEPTACLE_IMAGE_COORDS_X_KEY,
            self.RECEPTACLE_IMAGE_COORDS_Y_KEY,
            self.DEFAULT_IMG_COORD,
            "receptacleObjectId",
            **kwargs
        )

    def _get_object_vector(self, **kwargs) -> Dict:
        return self._get_vector(
            self.OBJECT_IMAGE_COORDS_X_KEY,
            self.OBJECT_IMAGE_COORDS_Y_KEY,
            self.DEFAULT_IMG_COORD,
            "objectId",
            **kwargs
        )

    def _validate_and_convert_params(self, **kwargs) -> Tuple[Action, Dict]:
        """Need a validation/conversion step for what ai2thor will accept as input
        to keep parameters more simple for the user (in this case, wrapping
        rotation degrees into an object)
        """
        action = Action(kwargs.get('action'))
        kwargs.pop('action')
        amount = self._get_amount(action, **kwargs)
        force = self._get_force(action, **kwargs)

        # TODO Consider the current "head tilt" value while validating the
        # input "horizon" value.
        horizon = kwargs.get(self.HORIZON_KEY, self.DEFAULT_HORIZON)
        rotation = kwargs.get(self.ROTATION_KEY, self.DEFAULT_ROTATION)

        rotation_vector = {'y': rotation}

        # Need to return a default value, Unity fails if returning None, or
        #  x: None, y: None for both vectors below
        object_vector = self._get_object_vector(**kwargs) if (
            action in OBJECT_IMAGE_ACTIONS) else self.DEFAULT_IMG_DICT

        receptacle_vector = self._get_receptacle_vector(**kwargs) if (
            action in RECEPTACLE_ACTIONS) else self.DEFAULT_IMG_DICT

        move_magnitude = self._get_move_magnitude(action, force, amount)
        (teleport_rotation, teleport_position) = self._get_teleport(**kwargs)
        clockwise = self._get_clockwise(**kwargs)
        (lateral, straight) = self._get_movement_direction(**kwargs)

        # TODO is this a feature we need?
        if self.config.is_noise_enabled():
            rotation = rotation * (1 + self._generate_noise())
            horizon = horizon * (1 + self._generate_noise())
            move_magnitude = move_magnitude * (1 + self._generate_noise())

        return action, dict(
            objectId=kwargs.get("objectId"),
            receptacleObjectId=kwargs.get("receptacleObjectId"),
            rotation=rotation_vector,
            horizon=horizon,
            teleportRotation=teleport_rotation,
            teleportPosition=teleport_position,
            moveMagnitude=move_magnitude,
            objectImageCoords=object_vector,
            receptacleObjectImageCoords=receptacle_vector,
            clockwise=clockwise,
            lateral=lateral,
            straight=straight
        )

    def _mcs_action_to_ai2thor_action(self, action: Action) -> str:
        if action == Action.CLOSE_OBJECT:
            # The AI2-THOR Python library has buggy error checking
            # specifically for the CloseObject action,
            # so just use our own custom action here.
            return "MCSCloseObject"
        elif action == Action.DROP_OBJECT:
            return "DropHandObject"
        elif action == Action.OPEN_OBJECT:
            # The AI2-THOR Python library has buggy error checking
            # specifically for the OpenObject action,
            # so just use our own custom action here.
            return "MCSOpenObject"

        return action.value

    def _generate_noise(self) -> float:
        """
        Returns a random value between -0.05 and 0.05 used to add noise to all
        numerical action parameters noise_enabled is True.
        Returns
        -------
        float
            A value between -0.05 and 0.05 (using random.uniform).
        """

        return random.uniform(self.MIN_NOISE, self.MAX_NOISE)



import unittest
from typing import Tuple

from shapely import geometry

import machine_common_sense as mcs


class TestReward(unittest.TestCase):

    def test_default_reward(self):
        goal = mcs.GoalMetadata()
        reward = mcs.Reward.calculate_reward(
            goal, objects=[{}], agent={}, number_steps=1,
            reach=1.0, steps_on_lava=0)
        self.assertEqual(reward, -mcs.reward.STEP_PENALTY)
        self.assertIsInstance(reward, float)

    def test_none_goal(self):
        goal = None
        reward = mcs.Reward.calculate_reward(
            goal, objects=[{}], agent={}, number_steps=5,
            reach=1.0, steps_on_lava=0)
        self.assertEqual(reward, -(5 * mcs.reward.STEP_PENALTY))
        self.assertIsInstance(reward, float)

    def test_penalty_step_calcuation(self):
        goal = None
        reward = mcs.Reward.calculate_reward(
            goal, objects=[{}], agent={}, number_steps=456,
            reach=1.0, steps_on_lava=0)
        penalty = 0 - (456 * mcs.reward.STEP_PENALTY)
        self.assertEqual(reward, penalty)
        self.assertIsInstance(reward, float)

    def test_steps_with_config_step_penalty(self):
        goal = None
        config_step_penalty = 0.25
        reward = mcs.Reward.calculate_reward(
            goal, objects=[{}], agent={}, number_steps=456,
            reach=1.0, steps_on_lava=0, step_penalty=config_step_penalty)
        penalty = 0 - (456 * config_step_penalty)
        self.assertEqual(reward, penalty)
        self.assertIsInstance(reward, float)

    def test_penalty_on_lava(self):
        goal = None
        reward = mcs.Reward.calculate_reward(
            goal, objects=[{}], agent={}, number_steps=1,
            reach=1.0, steps_on_lava=1)
        penalty = 0 - (1 * mcs.reward.STEP_PENALTY) - mcs.reward.LAVA_PENALTY
        self.assertEqual(reward, penalty)
        self.assertIsInstance(reward, float)

    def test_penalty_on_lava_multiple_steps(self):
        goal = None
        reward = mcs.Reward.calculate_reward(
            goal, objects=[{}], agent={}, number_steps=456,
            reach=1.0, steps_on_lava=2)
        penalty = 0 - (456 * mcs.reward.STEP_PENALTY) - \
            (2 * mcs.reward.LAVA_PENALTY)
        self.assertEqual(reward, penalty)
        self.assertIsInstance(reward, float)

    def test_penalty_on_lava_multiple_steps_with_config_lava_penalty(self):
        goal = None
        config_lava_penalty = 10.5
        reward = mcs.Reward.calculate_reward(
            goal, objects=[{}], agent={}, number_steps=456,
            reach=1.0, steps_on_lava=2, lava_penalty=config_lava_penalty)
        penalty = 0 - (456 * mcs.reward.STEP_PENALTY) - \
            (2 * config_lava_penalty)
        self.assertEqual(reward, penalty)
        self.assertIsInstance(reward, float)

    def test_steps_and_lava_config_penalty(self):
        goal = None
        config_lava_penalty = 10.5
        config_step_penalty = 0.25
        reward = mcs.Reward.calculate_reward(
            goal, objects=[{}], agent={}, number_steps=456,
            reach=1.0, steps_on_lava=2, lava_penalty=config_lava_penalty,
            step_penalty=config_step_penalty)
        penalty = 0 - (456 * config_step_penalty) - \
            (2 * config_lava_penalty)
        self.assertEqual(reward, penalty)
        self.assertIsInstance(reward, float)

    def test_steps_and_lava_config_is_none(self):
        current_score = 0
        config_lava_penalty = None
        config_step_penalty = None
        steps_on_lava = None
        reward = mcs.Reward._adjust_score_penalty(
            current_score=current_score, number_steps=456,
            steps_on_lava=steps_on_lava,
            lava_penalty=config_lava_penalty, step_penalty=config_step_penalty)
        penalty = current_score - (456 * mcs.reward.STEP_PENALTY) - \
            (0 * mcs.reward.LAVA_PENALTY)
        self.assertEqual(reward, penalty)
        self.assertIsInstance(reward, float)

    def test_steps_and_lava_config_penalty_score_is_not_one(self):
        current_score = 0
        config_lava_penalty = 10.5
        config_step_penalty = 0.25
        reward = mcs.Reward._adjust_score_penalty(
            current_score=current_score, number_steps=456, steps_on_lava=8,
            lava_penalty=config_lava_penalty, step_penalty=config_step_penalty)
        penalty = current_score - (456 * config_step_penalty) - \
            (8 * config_lava_penalty)
        self.assertEqual(reward, penalty)
        self.assertIsInstance(reward, float)

    def test_steps_and_lava_config_penalty_score_is_one(self):
        current_score = 1
        config_lava_penalty = 10.5
        config_step_penalty = 0.25
        reward = mcs.Reward._adjust_score_penalty(
            current_score=current_score, number_steps=456, steps_on_lava=8,
            lava_penalty=config_lava_penalty, step_penalty=config_step_penalty)
        penalty = current_score - ((456 - 1) * config_step_penalty)
        self.assertEqual(reward, penalty)
        self.assertIsInstance(reward, float)

    def test_target_in_empty_object_list(self):
        obj_list = []
        target_id = ''
        self.assertIsNone(
            mcs.Reward._Reward__get_object_from_list(
                obj_list, target_id))

    def test_target_in_object_list(self):
        obj_list = []
        for i in range(10):
            obj = {"objectId": str(i)}
            obj_list.append(obj)
        target_id = '7'
        results = mcs.Reward._Reward__get_object_from_list(
            obj_list, target_id)
        self.assertTrue(results)
        self.assertIsInstance(results, dict)
        self.assertEqual(results['objectId'], target_id)

    def test_target_not_in_object_list(self):
        obj_list = [{"objectId": str(i)} for i in range(10)]
        target_id = '111'  # not in list
        results = mcs.Reward._Reward__get_object_from_list(
            obj_list, target_id)
        self.assertIsNone(results)

    def test_duplicate_target_ids_in_list(self):
        obj_list = [{"objectId": str(i)} for i in range(10)]
        # add duplicate object
        obj_list.append({"objectId": '7', "duplicate": True})
        target_id = '7'
        results = mcs.Reward._Reward__get_object_from_list(
            obj_list, target_id)
        self.assertTrue(results)
        self.assertIsInstance(results, dict)
        self.assertEqual(results['objectId'], target_id)
        # expect that method still returns the first instance and not the
        # duplicate
        self.assertFalse('duplicate' in results)

    def test_convert_object_to_planar_polygon(self):
        goal_object = {'objectBounds': {'objectBoundsCorners': []}}
        # create lower plane (y = 0)
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 0.0, 'z': 0.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 1.0, 'y': 0.0, 'z': 0.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 1.0, 'y': 0.0, 'z': 1.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 0.0, 'z': 1.0})
        # create upper plane (y = 1)
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 1.0, 'z': 0.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 1.0, 'y': 1.0, 'z': 0.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 1.0, 'y': 1.0, 'z': 1.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 1.0, 'z': 1.0})

        polygon = mcs.Reward._convert_object_to_planar_polygon(goal_object)

        self.assertIsInstance(polygon, geometry.Polygon)
        self.assertEqual(len(list(polygon.exterior.coords)), 5)
        self.assertIsInstance(list(polygon.exterior.coords)[0], Tuple)

    def test_convert_skewed_object_to_planar_polygon(self):
        goal_object = {'objectBounds': {'objectBoundsCorners': []}}

        # create lower plane (y = 0)
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 0.0, 'z': 0.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 1.0, 'y': 0.0, 'z': 0.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 1.0, 'y': 0.0, 'z': 1.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 0.0, 'z': 1.0})
        # create upper plane (y = 1) but shifted over x and z by 1
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 1.0, 'y': 1.0, 'z': 1.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 2.0, 'y': 1.0, 'z': 1.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 2.0, 'y': 1.0, 'z': 2.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 1.0, 'y': 1.0, 'z': 2.0})

        polygon = mcs.Reward._convert_object_to_planar_polygon(goal_object)

        self.assertIsInstance(polygon, geometry.Polygon)
        self.assertEqual(len(list(polygon.exterior.coords)), 7)
        self.assertIsInstance(list(polygon.exterior.coords)[0], Tuple)

    def test_retrieval_reward(self):
        goal = mcs.GoalMetadata()
        goal.metadata['target'] = {'id': '0'}
        obj_list = []
        for i in range(10):
            obj = {"objectId": str(i), 'isPickedUp': not i}
            obj_list.append(obj)
        reward = mcs.Reward._calc_retrieval_reward(goal, obj_list, agent={},
                                                   performer_reach=1.0)
        self.assertEqual(reward, 1)
        self.assertIsInstance(reward, int)

    def test_retrieval_reward_nothing_pickedup(self):
        goal = mcs.GoalMetadata()
        goal.metadata['target'] = {'id': '0'}
        obj_list = []
        for i in range(10):
            obj = {"objectId": str(i), 'isPickedUp': False}
            obj_list.append(obj)
        reward = mcs.Reward._calc_retrieval_reward(goal, obj_list, agent={},
                                                   performer_reach=1.0)
        self.assertEqual(reward, 0)
        self.assertIsInstance(reward, int)

    def test_retrieval_reward_multi_target(self):
        goal = mcs.GoalMetadata()
        goal.metadata['targets'] = [{'id': '0'}, {'id': '1'}, {'id': '2'}]
        obj_list = []
        for i in range(10):
            obj_list.append({'objectId': str(i), 'isPickedUp': True})
        reward = mcs.Reward._calc_retrieval_reward(goal, obj_list, agent={},
                                                   performer_reach=1.0)
        self.assertEqual(reward, 1)
        self.assertIsInstance(reward, int)

    def test_retrieval_reward_multi_target_nothing_pickedup(self):
        goal = mcs.GoalMetadata()
        goal.metadata['targets'] = [{'id': '0'}, {'id': '1'}, {'id': '2'}]
        obj_list = []
        for i in range(10):
            obj_list.append({'objectId': str(i), 'isPickedUp': False})
        reward = mcs.Reward._calc_retrieval_reward(goal, obj_list, agent={},
                                                   performer_reach=1.0)
        self.assertEqual(reward, 0)
        self.assertIsInstance(reward, int)

    def test_retrieval_reward_multi_target_some_pickedup(self):
        goal = mcs.GoalMetadata()
        goal.metadata['targets'] = [{'id': '0'}, {'id': '1'}, {'id': '2'}]
        obj_list = []
        for i in range(10):
            obj_list.append({'objectId': str(i), 'isPickedUp': (i == 0)})
        reward = mcs.Reward._calc_retrieval_reward(goal, obj_list, agent={},
                                                   performer_reach=1.0)
        self.assertEqual(reward, 0)
        self.assertIsInstance(reward, int)

    def test_retrieval_reward_multi_target_more_pickedup(self):
        goal = mcs.GoalMetadata()
        goal.metadata['targets'] = [{'id': '0'}, {'id': '1'}, {'id': '2'}]
        obj_list = []
        for i in range(10):
            obj_list.append({'objectId': str(i), 'isPickedUp': (i != 0)})
        reward = mcs.Reward._calc_retrieval_reward(goal, obj_list, agent={},
                                                   performer_reach=1.0)
        self.assertEqual(reward, 0)
        self.assertIsInstance(reward, int)


if __name__ == '__main__':
    unittest.main()



from enum import Enum, unique


@unique
class ReturnStatus(Enum):
    CANNOT_ROTATE = "CANNOT_ROTATE"
    IS_CLOSED_COMPLETELY = "IS_CLOSED_COMPLETELY"
    IS_OPENED_COMPLETELY = "IS_OPENED_COMPLETELY"
    NOT_HELD = "NOT_HELD"
    NOT_INTERACTABLE = "NOT_INTERACTABLE"
    NOT_MOVEABLE = "NOT_MOVEABLE"
    NOT_OBJECT = "NOT_OBJECT"
    NOT_OPENABLE = "NOT_OPENABLE"
    NOT_RECEPTACLE = "NOT_RECEPTACLE"
    NOT_PICKUPABLE = "NOT_PICKUPABLE"
    NOT_VISIBLE = "NOT_VISIBLE"
    IS_LOCKED = "IS_LOCKED"
    OBSTRUCTED = "OBSTRUCTED"
    OUT_OF_REACH = "OUT_OF_REACH"
    NOT_AGENT = "NOT_AGENT"
    AGENT_IS_BUSY = "AGENT_IS_BUSY"
    AGENT_CURRENTLY_INTERACTING_WITH_PERFORMER = \
        "AGENT_CURRENTLY_INTERACTING_WITH_PERFORMER"
    SUCCESSFUL = "SUCCESSFUL"
    SUCCESSFUL_WITH_INVALID_PARAMETERS = "SUCCESSFUL_WITH_INVALID_PARAMETERS"
    UNDEFINED = "UNDEFINED"
    FAILED = "FAILED"



import unittest

import machine_common_sense as mcs
from machine_common_sense.config_manager import ConfigManager
from machine_common_sense.controller_logger import ControllerLogger
from machine_common_sense.controller_media import (
    DepthImageEventHandler, DepthVideoEventHandler, ImageVideoEventHandler,
    ObjectMaskImageEventHandler, SceneImageEventHandler,
    SegmentationVideoEventHandler, TopdownVideoEventHandler,
    UnityTopdownCameraCombinerEventHandler)
from machine_common_sense.history_writer import HistoryEventHandler
from machine_common_sense.subscriber import add_subscribers

from .mock_controller import MockController


class TestMCS(unittest.TestCase):

    def test_create_controller(self):
        # TODO How do we test this without starting the whole app?
        pass

    def test_add_subscribers(self):
        ctrl = MockController()
        cfg = ConfigManager({})
        ctrl._subscribers.clear()

        def return_true():
            return True

        def return_false():
            return False

        cfg.is_history_enabled = return_false
        cfg.is_save_debug_images = return_false
        cfg.is_depth_maps_enabled = return_false
        cfg.is_object_masks_enabled = return_false
        cfg.is_video_enabled = return_false
        add_subscribers(ctrl, cfg)
        self.assertEqual(len(ctrl._subscribers), 2)
        self.assertTrue(
            any([isinstance(subscriber, ControllerLogger)
                 for subscriber in ctrl._subscribers]))
        self.assertTrue(
            any([isinstance(subscriber, HistoryEventHandler)
                 for subscriber in ctrl._subscribers]))

        ctrl._subscribers.clear()
        cfg.is_save_debug_images = return_true
        cfg.is_depth_maps_enabled = return_false
        cfg.is_object_masks_enabled = return_false
        cfg.is_video_enabled = return_false
        add_subscribers(ctrl, cfg)
        self.assertEqual(len(ctrl._subscribers), 4)
        # the order of event handlers in the list doesn't matter
        # These assertions check that any of the elements in the list
        # are the correct handler instances
        self.assertTrue(any([isinstance(subscriber, DepthImageEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, SceneImageEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, ControllerLogger)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, HistoryEventHandler)
                             for subscriber in ctrl._subscribers]))

        ctrl._subscribers.clear()
        cfg.is_save_debug_images = return_false
        cfg.is_depth_maps_enabled = return_false
        cfg.is_object_masks_enabled = return_false
        cfg.is_video_enabled = return_true
        add_subscribers(ctrl, cfg)
        self.assertEqual(len(ctrl._subscribers), 4)
        self.assertTrue(any([isinstance(subscriber, ImageVideoEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber,
                                        UnityTopdownCameraCombinerEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, ControllerLogger)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, HistoryEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertFalse(any([isinstance(subscriber, TopdownVideoEventHandler)
                              for subscriber in ctrl._subscribers]))

        ctrl._subscribers.clear()
        cfg.is_save_debug_images = return_false
        cfg.is_depth_maps_enabled = return_false
        cfg.is_object_masks_enabled = return_true
        cfg.is_video_enabled = return_true
        add_subscribers(ctrl, cfg)
        self.assertEqual(len(ctrl._subscribers), 5)
        self.assertTrue(any([isinstance(subscriber, ImageVideoEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber,
                                        UnityTopdownCameraCombinerEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber,
                                        SegmentationVideoEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, ControllerLogger)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, HistoryEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertFalse(any([isinstance(subscriber, TopdownVideoEventHandler)
                              for subscriber in ctrl._subscribers]))

        ctrl._subscribers.clear()
        cfg.is_save_debug_images = return_true
        cfg.is_depth_maps_enabled = return_true
        cfg.is_object_masks_enabled = return_false
        cfg.is_video_enabled = return_false
        add_subscribers(ctrl, cfg)
        self.assertEqual(len(ctrl._subscribers), 4)
        self.assertTrue(any([isinstance(subscriber, DepthImageEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, SceneImageEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, ControllerLogger)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, HistoryEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertFalse(any([isinstance(subscriber, TopdownVideoEventHandler)
                              for subscriber in ctrl._subscribers]))

        ctrl._subscribers.clear()
        cfg.is_save_debug_images = return_true
        cfg.is_depth_maps_enabled = return_false
        cfg.is_object_masks_enabled = return_true
        cfg.is_video_enabled = return_false
        add_subscribers(ctrl, cfg)
        self.assertEqual(len(ctrl._subscribers), 5)
        self.assertTrue(any([isinstance(subscriber, DepthImageEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, SceneImageEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber,
                                        ObjectMaskImageEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, ControllerLogger)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, HistoryEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertFalse(any([isinstance(subscriber, TopdownVideoEventHandler)
                              for subscriber in ctrl._subscribers]))

        ctrl._subscribers.clear()
        cfg.is_save_debug_images = return_false
        cfg.is_depth_maps_enabled = return_true
        cfg.is_object_masks_enabled = return_true
        cfg.is_video_enabled = return_true
        add_subscribers(ctrl, cfg)
        self.assertEqual(len(ctrl._subscribers), 6)
        self.assertTrue(any([isinstance(subscriber, ImageVideoEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber,
                                        UnityTopdownCameraCombinerEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, DepthVideoEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber,
                                        SegmentationVideoEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, ControllerLogger)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, HistoryEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertFalse(any([isinstance(subscriber, TopdownVideoEventHandler)
                              for subscriber in ctrl._subscribers]))

        ctrl._subscribers.clear()
        cfg.is_save_debug_images = return_true
        cfg.is_depth_maps_enabled = return_true
        cfg.is_object_masks_enabled = return_true
        cfg.is_video_enabled = return_true
        add_subscribers(ctrl, cfg)
        self.assertEqual(len(ctrl._subscribers), 9)
        self.assertTrue(any([isinstance(subscriber, DepthImageEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, SceneImageEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber,
                                        ObjectMaskImageEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, ImageVideoEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber,
                                        UnityTopdownCameraCombinerEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, DepthVideoEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber,
                                        SegmentationVideoEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, ControllerLogger)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, HistoryEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertFalse(any([isinstance(subscriber, TopdownVideoEventHandler)
                              for subscriber in ctrl._subscribers]))

        ctrl._subscribers.clear()
        cfg.is_video_enabled = return_true
        cfg.is_top_down_plotter = return_true
        cfg.is_top_down_camera = return_false
        add_subscribers(ctrl, cfg)
        self.assertEqual(len(ctrl._subscribers), 9)
        self.assertTrue(any([isinstance(subscriber, DepthImageEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, SceneImageEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber,
                                        ObjectMaskImageEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, ImageVideoEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertFalse(any([isinstance(
            subscriber,
            UnityTopdownCameraCombinerEventHandler)
            for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, TopdownVideoEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, DepthVideoEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber,
                                        SegmentationVideoEventHandler)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, ControllerLogger)
                             for subscriber in ctrl._subscribers]))
        self.assertTrue(any([isinstance(subscriber, HistoryEventHandler)
                             for subscriber in ctrl._subscribers]))

    def test_load_scene_file_json(self):
        actual = mcs.load_scene_json_file("tests/test_scene.json")
        expected = {
            "ceilingMaterial": "Walls/WallDrywallWhite",
            "floorMaterial": "Fabrics/RUG4",
            "wallMaterial": "Walls/YellowDrywall",
            "objects": [{
                "id": "testBall",
                "type": "sphere",
                "materialFile": "Plastics/BlueRubber",
                "physics": True,
                "shows": [{
                    "stepBegin": 0,
                    "position": {
                        "x": 0,
                        "y": 0.5,
                        "z": 3
                    },
                    "scale": {
                        "x": 0.25,
                        "y": 0.25,
                        "z": 0.25
                    }
                }],
                "forces": [{
                    "stepBegin": 1,
                    "stepEnd": 1,
                    "vector": {
                        "x": 50,
                        "y": 0,
                        "z": 0
                    }
                }]
            }]
        }
        self.assertEqual(actual, expected)

    def test_load_scene_file_json_is_invalid(self):
        with self.assertRaises(ValueError):
            mcs.load_scene_json_file("tests/test_scene_invalid.json")

    def test_load_scene_file_json_is_missing(self):
        with self.assertRaises(FileNotFoundError):
            mcs.load_scene_json_file(
                "tests/test_scene_missing.json")


if __name__ == '__main__':
    unittest.main()



import json
import logging
import pathlib
import queue
import threading
import time
from abc import ABC, abstractmethod
from typing import Any

import cv2
import numpy as np
import PIL

logger = logging.getLogger(__name__)


class BaseRecorder(ABC):
    '''BaseRecorder class provides common functionality for all recorders.

    The abstract class handles the thread writing and item storage queue.
    '''

    _queue = None
    _thread = None
    num_recorded = 0
    recording = False

    def __init__(self, timeout: float = 1.0):
        self.timeout = timeout

    def start(self):
        '''Start the recorder.

        Prepares the queue for item storage and starts the write thread.
        '''
        logger.debug("Starting recorder")
        self.recording = True
        self._queue = queue.Queue()
        self._thread = threading.Thread(target=self.write, args=())
        self._thread.daemon = True
        self._thread.start()

    def flush(self) -> None:
        '''Write every item currently in the queue'''
        while not self._queue.empty():
            logger.debug("Flushing recorder queue")
            item = self._queue.get()
            self._write(item)
            self.num_recorded = self.num_recorded + 1

    def finish(self) -> None:
        '''Flush and close the recorder thread

        Further recordings should use a new recorder instance.
        '''
        logger.debug("Closing recorder")
        self.recording = False
        self.flush()
        self._thread.join()

    def write(self) -> None:
        '''Thread loop waiting for frames to enter the queue.'''
        if not self.recording:
            logger.warning("Recorder inactive. Unable to start thread.")
            return

        while True:
            if not self.recording:
                logger.debug("Recorder going inactive")
                return
            if not self._queue.empty():
                item = self._queue.get()
                self._write(item)
                self.num_recorded = self.num_recorded + 1
                logger.debug(f"Recorder wrote {self.num_recorded} items")
            else:
                time.sleep(self.timeout)

    def add(self, item: Any) -> None:
        '''Adds an item to the queue

        Args:
            item (Any): Item to be stored in queue

        Return:
            None
        '''
        if self.recording:
            logger.debug("Adding item to recorder queue")
            self._queue.put(item)
            logger.debug(f"Queue size is approximately {self._queue.qsize()}")

    @abstractmethod
    def _write(self):
        '''Subclasses to provide writing specifics'''
        pass


class JsonRecorder(BaseRecorder):
    '''Threaded json recorder'''

    def __init__(self, json_template: pathlib.Path,
                 timeout: float = 1.0, sort_keys=True, indent=4):
        '''Create the json recorder.

        Args:
            json_template(pathlib.Path): Filename template
            timeout (float): How often in seconds to check the queue

        Returns:
            None
        '''
        self.path = json_template
        self.sort_keys = sort_keys
        self.indent = indent
        super().__init__(timeout)
        super().start()

    def _write(self, data: dict) -> None:
        '''Save the json to disk

        Args:
            data (dict): The data to record

        Returns:
            None
        '''
        json_file = str(self.path).format(self.num_recorded)
        logger.debug(f"Writing {json_file}")
        with open(json_file, 'w') as json_file:
            json.dump(
                data,
                json_file,
                sort_keys=self.sort_keys,
                indent=self.indent)


class ImageRecorder(BaseRecorder):
    '''Threaded image recorder

    '''

    def __init__(self,
                 img_template: pathlib.Path,
                 timeout: float = 1.0):
        '''Create the image recorder.

        Args:
            img_template(pathlib.Path): Filename template
            timeout (float): How often in seconds to check the queue

        Returns:
            None
        '''
        self.path = img_template
        super().__init__(timeout)
        super().start()

    def _write(self, image: PIL.Image.Image) -> None:
        '''Save the image to disk

        Args:
            image (Pil.Image.Image): The image to record

        Returns:
            None
        '''
        image_filename = str(self.path).format(self.num_recorded)
        logger.debug(f"Writing {image_filename}")
        image.save(image_filename)


class VideoRecorder(BaseRecorder):
    '''Threaded video recorder'''

    def __init__(self,
                 vid_path: pathlib.Path,
                 fps: int,
                 fourcc: str = 'mp4v',
                 timeout: float = 1.0):
        '''Create the video recorder.

        Args:
            vid_path (pathlib.path): the output video file path
            fps (int): video frame rate per second
            fourcc (str): opencv fourcc / codec string
            timeout (float): thread sleep timeout in seconds

        Returns:
            None
        '''
        self.path = vid_path
        self._frames_written = 0
        self.fourcc = fourcc
        self.fps = fps
        self.writer = None
        super().__init__(timeout)
        super().start()

    def add(self, frame: PIL.Image.Image) -> None:
        '''Adds the video frame to the queue.

        Requires that the start function was called
        otherwise the frame is ignored.

        Args:
            frame (PIL.Image.Image): RGB video frame to be written

        Returns:
            None

        Raises:
            ValueError: If frame is a different size from the initial
        '''

        if self.writer is None:
            self.width, self.height = frame.size
            logger.debug(
                f"Establishing video writer size"
                f"({self.width},{self.height}) from first frame")
            self.writer = cv2.VideoWriter(str(self.path),
                                          cv2.VideoWriter_fourcc(*self.fourcc),
                                          self.fps,
                                          (self.width, self.height),
                                          True)
        width, height = frame.size
        if (width, height) != (self.width, self.height):
            raise ValueError(f"Wrong size frame ({width}, {height}) for "
                             f"video writer ({self.width}, {self.height})")

        super().add(frame)

    def _write(self, frame: PIL.Image.Image) -> None:
        '''Write the frame to the video recording

        Args:
            frame (PIL.Image.Image): Image frame to be written

        Return:
            None
        '''
        # convert BGR PIL image to RGB for opencv video writer
        cv_frame = np.array(frame.convert('RGB'))[:, :, ::-1]
        logger.debug(f"Writing frame #{self.num_recorded} to {self.path}")
        self.writer.write(cv_frame)

    def finish(self) -> None:
        '''Deactivate the recorder so that it does not accept more frames.

        Frames that remain in the buffer will be written and the recorder
        closed.
        '''
        logger.debug("Closing video recorder")
        super().finish()
        if self.writer:
            self.writer.release()



import unittest

import numpy as np
import PIL

from machine_common_sense.goal_metadata import GoalMetadata
from machine_common_sense.object_metadata import ObjectMetadata
from machine_common_sense.serializer import ISerializer, SerializerMsgPack
from machine_common_sense.step_metadata import StepMetadata


class TestISerializer(unittest.TestCase):

    def test_version(self):
        version = "0.1.0"
        self.assertEqual(ISerializer.version(), version)

    def test_serialize(self):
        with self.assertRaises(NotImplementedError):
            ISerializer.serialize(None)

    def test_deserialize(self):
        with self.assertRaises(NotImplementedError):
            ISerializer.deserialize(None)

    def test_image_to_bytes(self):
        size = (50, 100)
        black = (0, 0, 0)
        img = PIL.Image.new("RGB", size=size, color=black)
        image_bytes = ISerializer.image_to_bytes(img)
        self.assertIsInstance(image_bytes, bytes)
        image_str = image_bytes.decode('ISO-8859-1')
        self.assertTrue("PNG" in image_str)
        self.assertTrue("IHDR" in image_str)
        self.assertTrue("IDAT" in image_str)
        self.assertTrue("IEND" in image_str)

    def test_bytes_to_image(self):
        # bytes representing a black (0,0,0) 50x100 image
        image_bytes = (
            b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x002\x00'
            b'\x00\x00d\x08\x02\x00\x00\x00K\x8a\xf4\x0c\x00\x00\x00'
            b'&IDATx\x9c\xed\xc11\x01\x00\x00\x00\xc2\xa0\xf5Om\r\x0f'
            b'\xa0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\xe0\xc3\x00:\xfc\x00\x01\xc0gM\x95\x00\x00\x00'
            b'\x00IEND\xaeB`\x82')
        image = ISerializer.bytes_to_image(image_bytes)
        self.assertIsInstance(image, PIL.Image.Image)
        self.assertEqual(image.size, (50, 100))
        self.assertTrue(all(pixel == (0, 0, 0) for pixel in image.getdata()))


class TestSerializerMsgPack(unittest.TestCase):

    # allow larger unit tests comparisons for large dictionaries
    maxDiff = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.serializer = SerializerMsgPack()

    @classmethod
    def tearDownClass(cls) -> None:
        return super().tearDownClass()

    def test_serialize(self):
        packed_bytes = self.serializer.serialize(StepMetadata())
        self.assertIsInstance(packed_bytes, bytes)

    def test_deserializer(self):  # sourcery skip: class-extract-method
        # serialize a default StepMetadata()
        step_metadata = StepMetadata()
        packed_bytes = self.serializer.serialize(step_metadata)
        self.assertIsInstance(packed_bytes, bytes)

        # deserialize bytes back to a StepMetadata
        unpacked_step_metadata = self.serializer.deserialize(packed_bytes)
        self.assertIsInstance(unpacked_step_metadata, StepMetadata)

        # compare the two default StepMetadata objects
        self.assertEqual(
            unpacked_step_metadata.goal.__dict__,
            step_metadata.goal.__dict__)
        del step_metadata.goal
        del unpacked_step_metadata.goal
        self.assertEqual(
            unpacked_step_metadata.__dict__,
            step_metadata.__dict__)

    def test_ext_pack_with_stepmetadata(self):
        step_metadata = StepMetadata()
        step_exttype = self.serializer._ext_pack(step_metadata)
        self.assertEqual(step_exttype.code, 1)
        self.assertIsInstance(step_exttype.data, bytes)

    def test_ext_pack_with_tuple(self):
        test_tuple = (1, 'a')
        tuple_exttype = self.serializer._ext_pack(test_tuple)
        self.assertEqual(tuple_exttype.code, 2)
        self.assertIsInstance(tuple_exttype.data, bytes)
        self.assertEqual(tuple_exttype.data, b'\x92\x01\xa1a')

    def test_ext_pack_with_image(self):
        size = (50, 100)
        black = (0, 0, 0)
        img = PIL.Image.new("RGB", size=size, color=black)
        image_exttype = self.serializer._ext_pack(img)
        self.assertEqual(image_exttype.code, 3)
        self.assertIsInstance(image_exttype.data, bytes)
        image_str = image_exttype.data.decode('ISO-8859-1')
        self.assertTrue("PNG" in image_str)
        self.assertTrue("IHDR" in image_str)
        self.assertTrue("IDAT" in image_str)
        self.assertTrue("IEND" in image_str)

    def test_ext_pack_with_goalmetadata(self):
        goal_md = GoalMetadata()
        goal_exttype = self.serializer._ext_pack(goal_md)
        self.assertEqual(goal_exttype.code, 4)
        self.assertIsInstance(goal_exttype.data, bytes)

    def test_ext_pack_with_objectmetadata(self):
        obj_md = ObjectMetadata()
        obj_exttype = self.serializer._ext_pack(obj_md)
        self.assertEqual(obj_exttype.code, 5)
        self.assertIsInstance(obj_exttype.data, bytes)

    def test_ext_pack_with_ndarray(self):
        array = np.zeros(10)
        np_exttype = self.serializer._ext_pack(array)
        self.assertEqual(np_exttype.code, 6)
        self.assertIsInstance(np_exttype.data, bytes)

    def test_ext_unpack_with_stepmetadata(self):
        # from default StepMetadata()
        test_data = (
            b'\xdc\x00\x11\x90\xc7\x13\x02\x92\xcb\x00\x00\x00\x00'
            b'\x00\x00\x00\x00\xcb\x00\x00\x00\x00\x00\x00\x00\x00\xc7\x13'
            b'\x02\x92\xcb\x00\x00\x00\x00\x00\x00\x00\x00\xcb\x00\x00\x00'
            b'\x00\x00\x00\x00\x00\xcb\x00\x00\x00\x00\x00\x00\x00\x00\xcb'
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x90\xd7\x04\x97\xc0\xa0\xa0'
            b'\x00\x00\xc0\x80\xcb\x00\x00\x00\x00\x00\x00\x00\x00\x90\x90'
            b'\x90\x80\xa9UNDEFINED\x00\xcb\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x90')
        unpacked_step = self.serializer._ext_unpack(code=1, data=test_data)
        self.assertIsInstance(unpacked_step, StepMetadata)
        step_metadata = StepMetadata()
        # Removing default goal metadata as they don't compare well
        # as nested dictionaries with differing ids. Individual goal elements
        # are the same however.
        self.assertEqual(
            unpacked_step.goal.__dict__,
            step_metadata.goal.__dict__)
        del step_metadata.goal
        del unpacked_step.goal
        self.assertEqual(unpacked_step.__dict__, step_metadata.__dict__)

    def test_ext_unpack_with_tuple(self):
        test_tuple = (1, 'a')
        test_data = b'\x92\x01\xa1a'
        unpacked_tuple = self.serializer._ext_unpack(code=2, data=test_data)
        self.assertIsInstance(unpacked_tuple, tuple)
        self.assertEqual(unpacked_tuple, test_tuple)

    def test_ext_unpack_with_image(self):
        # from (50,100) image of all black pixels (0,0,0)
        test_data = (
            b'\xc4_\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x002'
            b'\x00\x00\x00d\x08\x02\x00\x00\x00K\x8a\xf4\x0c\x00\x00\x00'
            b'&IDATx\x9c\xed\xc11\x01\x00\x00\x00\xc2\xa0\xf5Om\r\x0f'
            b'\xa0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\xe0\xc3\x00:\xfc\x00\x01\xc0gM\x95\x00\x00\x00'
            b'\x00IEND\xaeB`\x82')
        unpacked_image = self.serializer._ext_unpack(3, test_data)
        self.assertIsInstance(unpacked_image, PIL.Image.Image)
        self.assertEqual(unpacked_image.size, (50, 100))
        self.assertTrue(all(pixel == (0, 0, 0)
                            for pixel in unpacked_image.getdata()))

    def test_ext_unpack_with_goalmetadata(self):
        # from default GoalMetadata()
        test_data = b'\x97\xc0\xa0\xa0\x00\x00\xc0\x80'
        unpacked_goal = self.serializer._ext_unpack(4, test_data)
        self.assertIsInstance(unpacked_goal, GoalMetadata)
        self.assertEqual(unpacked_goal.__dict__, GoalMetadata().__dict__)

    def test_ext_unpack_with_objectmetadata(self):
        # from default ObjectMetadata()
        obj_data = (
            b'\xdc\x00\x10\xa0\x90\x80\xcb\xbf\xf0\x00\x00\x00\x00'
            b'\x00\x00\xcb\xbf\xf0\x00\x00\x00\x00\x00\x00\xcb\xbf\xf0\x00'
            b'\x00\x00\x00\x00\x00\xc2\xcb\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x90\x80\x80\x80\xa0\x90\x90\xc2')
        unpacked_objectmetadata = self.serializer._ext_unpack(
            code=5, data=obj_data)
        self.assertIsInstance(unpacked_objectmetadata, ObjectMetadata)
        self.assertEqual(
            unpacked_objectmetadata.__dict__,
            ObjectMetadata().__dict__)

    def test_ext_unpack_with_ndarray(self):
        # from [0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
        test_data = (
            b'\x9a\xcb\x00\x00\x00\x00\x00\x00\x00\x00\xcb\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\xcb\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\xcb\x00\x00\x00\x00\x00\x00\x00\x00\xcb\x00\x00\x00\x00\x00'
            b'\x00\x00\x00\xcb\x00\x00\x00\x00\x00\x00\x00\x00\xcb\x00\x00'
            b'\x00\x00\x00\x00\x00\x00\xcb\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\xcb\x00\x00\x00\x00\x00\x00\x00\x00\xcb\x00\x00\x00\x00\x00'
            b'\x00\x00\x00')
        unpacked_ndarray = self.serializer._ext_unpack(6, test_data)
        self.assertIsInstance(unpacked_ndarray, np.ndarray)
        self.assertFalse(np.all(unpacked_ndarray))
        self.assertEqual(unpacked_ndarray.size, 10)


if __name__ == '__main__':
    unittest.main()



import ast
import configparser  # noqa: F401
import logging
import os
from enum import Enum, unique
from typing import Dict, List, Optional, Tuple

import numpy as np
from pydantic import BaseModel as PydanticBaseModel

from .action import Action
from .goal_metadata import GoalCategory, GoalMetadata

logger = logging.getLogger(__name__)


def to_camel_case(string: str) -> str:
    '''Converts a snake case string to camel case'''
    words = string.split('_')
    return ''.join(word.capitalize() if word !=
                   words[0] else word for word in words)


class BaseModel(PydanticBaseModel):
    # global configs are enabled by creating a custom BaseModel class
    # contain the desired config
    class Config:
        alias_generator = to_camel_case
        allow_population_by_field_name = True
        arbitrary_types_allowed = True


@unique
class MetadataTier(Enum):
    # Normal metadata plus metadata for all hidden objects
    ORACLE = 'oracle'
    # No metadata, except for the images, depth masks, object masks,
    # and haptic/audio feedback
    LEVEL_2 = 'level2'
    # No metadata, except for the images, depth masks, and haptic/audio
    # feedback
    LEVEL_1 = 'level1'
    # No metadata, except for the images and haptic/audio
    # feedback
    NONE = 'none'
    # Default metadata level if none specified, meant for use during
    # development
    DEFAULT = 'default'


class Vector3d(BaseModel):
    x: Optional[float] = 0.0
    y: Optional[float] = 0.0
    z: Optional[float] = 0.0


class GoalObject(BaseModel):
    id: str
    image: bytes = None


class Goal(BaseModel):
    action_list: List = None  # TODO fields.Raw()
    answer: Optional[Dict]
    category: Optional[str]
    description: Optional[str]
    domains_info: Optional[Dict]
    habituation_total: Optional[int]
    info_list: Optional[List[str]]
    last_preview_phase_step: Optional[int]
    last_step: Optional[int]
    metadata: Optional[Dict]
    objects_info: Optional[Dict]
    scene_info: Optional[Dict]
    skip_preview_phase: Optional[bool]
    task_list: List[str] = None
    type_list: List[str] = None
    triggered_by_target_sequence: Optional[List[str]]


class ActionConfig(BaseModel):
    step_begin: int
    step_end: Optional[int]
    id: str
    is_loop_animation: Optional[bool] = False


class SequenceConfig(BaseModel):
    animation: str
    end_point: Vector3d


class AgentMovementConfig(BaseModel):
    step_begin: int
    repeat: bool
    sequence: List[SequenceConfig] = []


class AgentSettings(BaseModel):
    chest: int = -1
    chest_material: int = -1
    eyes: int = -1
    feet: int = -1
    feet_material: int = -1
    glasses: int = -1
    hair: int = -1
    hair_material: int = -1
    hat_material: int = -1
    hide_hair: bool = False
    is_elder: bool = False
    jacket: int = -1
    jacket_material: int = -1
    legs: int = -1
    legs_material: int = -1
    show_beard: bool = False
    show_glasses: bool = False
    show_jacket: bool = False
    show_tie: bool = False
    skin: int = -1
    tie: int = -1
    tie_material: int = -1


class ChangeMaterialConfig(BaseModel):
    step_begin: int
    materials: List[str]


class ForceConfig(BaseModel):
    step_begin: int
    step_end: int
    vector: Vector3d = Vector3d(x=0, y=0, z=0)
    impulse: bool = False
    relative: bool = False
    repeat: bool = False
    step_wait: Optional[int]


class MoveConfig(BaseModel):
    step_begin: int
    step_end: int
    vector: Vector3d = Vector3d(x=0, y=0, z=0)
    repeat: bool = False
    step_wait: int = 0
    global_space: bool = False


class OpenCloseConfig(BaseModel):
    step: int
    open: bool


class PhysicsConfig(BaseModel):
    enable: bool = False
    angular_drag: float
    bounciness: float = None
    drag: float = None
    dynamic_friction: float
    static_friction: float


class LipsGapSpanConfig(BaseModel):
    low: float
    high: float


class LipGapsConfig(BaseModel):
    front: Optional[List[LipsGapSpanConfig]]
    back: Optional[List[LipsGapSpanConfig]]
    left: Optional[List[LipsGapSpanConfig]]
    right: Optional[List[LipsGapSpanConfig]]


class PlatformLipsConfig(BaseModel):
    front: bool = False
    back: bool = False
    left: bool = False
    right: bool = False
    gaps: Optional[LipGapsConfig]


class Vector2dInt(BaseModel):
    x: Optional[int]
    z: Optional[int]


class FloorPartitionConfig(BaseModel):
    left_half: Optional[float]
    right_half: Optional[float]


class FloorTexturesConfig(BaseModel):
    material: str
    positions: List[Vector2dInt] = []


class ShowConfig(BaseModel):
    step_begin: int
    position: Vector3d = Vector3d(x=0, y=0, z=0)
    rotation: Vector3d = Vector3d(x=0, y=0, z=0)
    scale: Vector3d = Vector3d(x=1, y=1, z=1)
    bounding_box: Optional[List[Dict]]  # debug property


class SizeConfig(BaseModel):
    step_begin: int
    step_end: int
    size: Vector3d = Vector3d(x=1, y=1, z=1)
    repeat: bool = False
    step_wait: int = 0


class SingleStepConfig(BaseModel):
    step_begin: int


class StepBeginEndConfig(BaseModel):
    step_begin: int
    step_end: int
    repeat: bool = False
    step_wait: int = 0


class TeleportConfig(BaseModel):
    step_begin: int
    position: Vector3d = Vector3d(x=0, y=0, z=0)


class TransformConfig(BaseModel):
    position: Vector3d = Vector3d(x=0, y=0, z=0)
    rotation: Vector3d = Vector3d(x=0, y=0, z=0)
    scale: Vector3d = Vector3d(x=1, y=1, z=1)


class RoomMaterials(BaseModel):
    front: Optional[str]
    left: Optional[str]
    right: Optional[str]
    back: Optional[str]


class PerformerStart(BaseModel):
    position: Vector3d
    rotation: Vector3d


class LidConfig(BaseModel):
    step_begin: int
    lid_attachment_obj_id: str


class SceneObject(BaseModel):
    id: str
    type: str  # should this be an enum?
    actions: Optional[List[ActionConfig]]
    agent_movement: Optional[AgentMovementConfig]
    agent_settings: Optional[AgentSettings]
    associated_with_agent: Optional[str] = ""
    center_of_mass: Optional[Vector3d]
    change_materials: Optional[List[ChangeMaterialConfig]]
    debug: dict = None
    forces: List[ForceConfig] = None
    ghosts: List[StepBeginEndConfig] = None
    hides: List[SingleStepConfig] = None
    kinematic: Optional[bool]
    lid_attachment: Optional[LidConfig]
    location_parent: Optional[str]
    locked: bool = False
    mass: Optional[float]
    materials: List[str] = None
    # deprecated; please use materials
    material_file: Optional[str]
    max_angular_velocity: Optional[float]
    # Docs say moveable's default is dependant on type.  That could
    # be a problem for the concrete classes.  Needs more review later
    moveable: Optional[bool]
    moves: List[MoveConfig] = None
    null_parent: Optional[TransformConfig]
    openable: Optional[bool]
    opened: Optional[bool]
    open_close: List[OpenCloseConfig] = None
    lips: Optional[PlatformLipsConfig]
    physics: Optional[bool]
    physics_properties: Optional[PhysicsConfig]
    pickupable: Optional[bool]
    receptacle: Optional[bool]
    reset_center_of_mass: Optional[bool]
    reset_center_of_mass_at_y: Optional[float]
    resizes: List[SizeConfig] = None
    rotates: List[MoveConfig] = None
    salient_materials: List[str] = None
    seesaw: Optional[bool]
    shows: List[ShowConfig] = None
    shrouds: List[StepBeginEndConfig] = None
    states: List[List[str]] = None
    structure: Optional[bool]
    teleports: List[TeleportConfig] = None
    triggered_by: Optional[bool]
    toggle_physics: List[SingleStepConfig] = None
    torques: List[ForceConfig] = None

    # These are deprecated, but needed for Eval 3 backwards compatibility
    can_contain_target: Optional[bool]
    obstacle: Optional[bool]
    occluder: Optional[bool]
    position_y: Optional[float]
    stack_target: Optional[bool]


class ConfigManager:

    DEFAULT_ROOM_DIMENSIONS = Vector3d(x=10, y=3, z=10)
    CONFIG_FILE_ENV_VAR = 'MCS_CONFIG_FILE_PATH'
    CONFIG_DEFAULT_SECTION = 'MCS'
    CONFIG_EVALUATION_NAME = 'evaluation_name'
    CONFIG_HISTORY_ENABLED = 'history_enabled'
    CONFIG_METADATA_TIER = 'metadata'
    CONFIG_NOISE_ENABLED = 'noise_enabled'
    CONFIG_SAVE_DEBUG_IMAGES = 'save_debug_images'
    CONFIG_SAVE_DEBUG_JSON = 'save_debug_json'
    CONFIG_SIZE = 'size'
    CONFIG_TEAM = 'team'
    CONFIG_VIDEO_ENABLED = 'video_enabled'
    CONFIG_LAVA_PENALTY = 'lava_penalty'
    CONFIG_STEPS_ALLOWED_IN_LAVA = 'steps_allowed_in_lava'
    CONFIG_STEP_PENALTY = 'step_penalty'
    CONFIG_GOAL_REWARD = 'goal_reward'
    CONFIG_TOP_DOWN_PLOTTER = 'top_down_plotter'
    CONFIG_TOP_DOWN_CAMERA = 'top_down_camera'
    CONFIG_TIMEOUT = 'timeout'

    # Please keep the aspect ratio as 3:2 because the IntPhys scenes are built
    # on this assumption.
    SCREEN_WIDTH_DEFAULT = 600
    SCREEN_WIDTH_MIN = 450

    # Default steps allowed in lava before calling end scene
    STEPS_ALLOWED_IN_LAVA_DEFAULT = 0

    # Default time to allow on a single step before timing out
    # is 1 hour (represented in seconds)
    TIMEOUT_DEFAULT = 3600

    def __init__(self, config_file_or_dict=None):
        '''
        Configuration preferences passed in by the user.
        '''
        self._config = configparser.ConfigParser()

        # For config, look for environment variable first,
        # then look at config_file_or_dict from constructor
        try:
            if (os.getenv(self.CONFIG_FILE_ENV_VAR) is not None):
                self._read_in_config_file(os.getenv(self.CONFIG_FILE_ENV_VAR))
            elif (isinstance(config_file_or_dict, dict)):
                self._read_in_config_dict(config_file_or_dict)
            elif(isinstance(config_file_or_dict, str)):
                self._read_in_config_file(config_file_or_dict)
            else:
                raise FileNotFoundError("No config options given")
        except FileNotFoundError as err:
            raise RuntimeError("Configuration not set") from err

        self._validate_screen_size()

    def _read_in_config_dict(self, config_dict):
        self._config[self.CONFIG_DEFAULT_SECTION] = config_dict
        logger.info('No config file given or file path does not exist,'
                    ' using config dictionary')
        logger.info(f"Read in config dictionary: {str(config_dict)}")

    def _read_in_config_file(self, config_file_path):
        if os.path.exists(config_file_path):
            self._config.read(config_file_path)
            logger.info(f"Config File Path: {config_file_path}")
        else:
            logger.warning(f"No config file at given path: {config_file_path}")
            raise FileNotFoundError()

    def _validate_screen_size(self):
        if(self.get_size() < self.SCREEN_WIDTH_MIN):
            self._config.set(
                self.CONFIG_DEFAULT_SECTION,
                self.CONFIG_SIZE,
                str(self.SCREEN_WIDTH_DEFAULT)
            )

    def is_file_writing_enabled(self):
        return (
            self.is_save_debug_images() or
            self.is_save_debug_json() or
            self.is_video_enabled()
        )

    def get_evaluation_name(self):
        return self._config.get(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_EVALUATION_NAME,
            fallback=''
        )

    def get_metadata_tier(self):
        metadata = self._config.get(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_METADATA_TIER,
            fallback='default'
        )

        return MetadataTier(metadata)

    def set_metadata_tier(self, mode):
        self._config.set(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_METADATA_TIER,
            mode
        )

    def get_size(self):
        return self._config.getint(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_SIZE,
            fallback=self.SCREEN_WIDTH_DEFAULT
        )

    def get_team(self):
        return self._config.get(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_TEAM,
            fallback=''
        )

    def is_history_enabled(self):
        return self._config.getboolean(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_HISTORY_ENABLED,
            fallback=True
        )

    def is_noise_enabled(self):
        return self._config.getboolean(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_NOISE_ENABLED,
            fallback=False
        )

    def is_save_debug_images(self):
        return self._config.getboolean(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_SAVE_DEBUG_IMAGES,
            fallback=False
        )

    def is_save_debug_json(self):
        return self._config.getboolean(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_SAVE_DEBUG_JSON,
            fallback=False
        )

    def is_video_enabled(self):
        return self._config.getboolean(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_VIDEO_ENABLED,
            fallback=False
        )

    def is_depth_maps_enabled(self) -> bool:
        metadata_tier = self.get_metadata_tier()
        return metadata_tier in [
            MetadataTier.LEVEL_1,
            MetadataTier.LEVEL_2,
            MetadataTier.ORACLE,
        ]

    def is_object_masks_enabled(self) -> bool:
        metadata_tier = self.get_metadata_tier()
        return (
            metadata_tier != MetadataTier.LEVEL_1 and
            metadata_tier in
            [
                MetadataTier.LEVEL_2,
                MetadataTier.ORACLE,
            ]
        )

    def get_screen_size(self) -> Tuple[int, int]:
        return (self.get_screen_width(), self.get_screen_height())

    def get_screen_width(self) -> int:
        return self.get_size()

    def get_screen_height(self) -> int:
        size = self.get_size()
        return int(size / 3 * 2)

    def get_lava_penalty(self):
        return self._config.getfloat(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_LAVA_PENALTY,
            fallback=None
        )

    def get_step_penalty(self):
        return self._config.getfloat(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_STEP_PENALTY,
            fallback=None
        )

    def get_goal_reward(self):
        return self._config.getfloat(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_GOAL_REWARD,
            fallback=None
        )

    def get_steps_allowed_in_lava(self):
        return self._config.getint(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_STEPS_ALLOWED_IN_LAVA,
            fallback=self.STEPS_ALLOWED_IN_LAVA_DEFAULT
        )

    def get_timeout(self):
        """ Time (in seconds) to allow a run to be idle
        before attempting to end scene"""
        return self._config.getint(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_TIMEOUT,
            fallback=self.TIMEOUT_DEFAULT
        )

    def is_top_down_plotter(self) -> bool:
        """Toggles whether old plotter should be used to create top down
        videos if videos are enabled."""
        return self._config.getboolean(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_TOP_DOWN_PLOTTER,
            fallback=False
        )

    def is_top_down_camera(self) -> bool:
        """Toggles whether the new top down camera is used to create top down
        videos if videos are enabled."""
        return self._config.getboolean(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_TOP_DOWN_CAMERA,
            fallback=True
        )


class SceneConfiguration(BaseModel):
    '''Class for keeping track of scene configuration'''
    ceiling_material: Optional[str]
    debug: Optional[Dict]
    floor_material: Optional[str]
    floor_properties: Optional[PhysicsConfig]
    floor_textures: List[FloorTexturesConfig] = []
    goal: Optional[Goal]
    holes: List[Vector2dInt] = []
    intuitive_physics: bool = False
    isometric: bool = False
    lava: List[Vector2dInt] = []
    name: Optional[str]
    objects: List[SceneObject] = []
    observation: bool = False  # deprecated; please use intuitivePhysics
    partition_floor: Optional[FloorPartitionConfig]
    performer_start: Optional[PerformerStart]
    restrict_open_doors: Optional[bool]
    restrict_open_objects: Optional[bool]
    room_dimensions: Vector3d = ConfigManager.DEFAULT_ROOM_DIMENSIONS
    room_materials: Optional[RoomMaterials]
    screenshot: bool = False  # developer use only; for the image generator
    version: Optional[int]
    wall_material: Optional[str]
    wall_properties: Optional[PhysicsConfig]
    toggle_lights: List[StepBeginEndConfig] = []

    # These are deprecated, but needed for Eval 3 backwards compatibility
    evaluation: Optional[str]
    evaluation_only: Optional[bool]
    eval_name: Optional[str]
    hypercube_number: Optional[int]
    scene_number: Optional[int]
    sequence_number: Optional[int]
    training: Optional[bool]

    def is_passive_scene(self) -> bool:
        """Return whether this scene is a passive scene."""
        goal = self.goal or Goal()
        # Passive physics scenes will have the intuitive_physics property and
        # the INTUITIVE_PHYSICS goal category; passive agent (NYU) scenes will
        # have the isometric property and the AGENTS goal category; other
        # passive scenes will have the PASSIVE goal category.
        return self.intuitive_physics or self.isometric or goal.category in [
            GoalCategory.AGENTS.value,
            GoalCategory.INTUITIVE_PHYSICS.value,
            GoalCategory.PASSIVE.value
        ]

    """
    @post_dump
    def remove_none(self, d, **kwargs) -> Dict:
        '''Remove all none's from dictionaries'''
        for key, value in dict(d).items():
            if isinstance(value, dict):
                d[key] = self.remove_none(value)
            if isinstance(value, list):
                for index, val in enumerate(value):
                    if isinstance(val, dict):
                        value[index] = self.remove_none(val)
            if value is None:
                del d[key]
        return d
    """

    def retrieve_object_states(self,
                               object_id, step_number):
        """Return the state list at the current step for the object with the
        given ID from the scene configuration data, if any."""
        state_list_each_step = next(
            (
                object_config.states or []
                for object_config in self.objects
                if object_config.id == object_id
            ),
            [],
        )

        # Retrieve the object's states in the current step.
        if len(state_list_each_step) > step_number:
            state_list = state_list_each_step[step_number]
            # Validate the data type.
            if state_list is not None:
                if not isinstance(state_list, list):
                    return [state_list]
                return [str(state) for state in state_list]
        return []

    def retrieve_goal(self, steps_allowed_in_lava=0):
        if not self.goal:
            return self.update_goal_target_image(GoalMetadata(
                steps_allowed_in_lava=steps_allowed_in_lava
            ))

        goal = self.goal

        # Transform action list data from strings to tuples.
        action_list = goal.action_list or []
        for index, action_list_at_step in enumerate(action_list):
            action_list[index] = [
                Action.input_to_action_and_params(action)
                if isinstance(action, str) else action
                for action in action_list_at_step
            ]

        if goal.category:
            # Backwards compatibility
            goal.metadata['category'] = goal.category

        return self.update_goal_target_image(
            GoalMetadata(
                action_list=action_list or None,
                category=goal.category or '',
                description=goal.description or '',
                habituation_total=goal.habituation_total or 0,
                last_preview_phase_step=(goal.last_preview_phase_step or 0),
                last_step=goal.last_step or None,
                metadata=goal.metadata or {},
                steps_allowed_in_lava=steps_allowed_in_lava,
                triggered_by_target_sequence=goal.triggered_by_target_sequence or None  # noqa
            )
        )

    def retrieve_lava(self) -> List[Tuple[float, float, float, float]]:
        """Return the list of lava locations as (X1, Z1, X2, Z2) tuples, where
        X1/Z1 is the top-left corner and X2/Z2 is the bottom-right corner."""
        lava = [
            (area.x - 0.5, area.z - 0.5, area.x + 0.5, area.z + 0.5)
            for area in self.lava
        ]
        if self.partition_floor and (
            self.partition_floor.left_half or
            self.partition_floor.right_half
        ):
            x_half = self.room_dimensions.x / 2.0
            z_half = self.room_dimensions.z / 2.0
            x_left_scale = x_half * min(self.partition_floor.left_half, 1)
            lava.append((-x_half, z_half, (-x_half + x_left_scale), -z_half))
            x_right_scale = x_half * min(self.partition_floor.right_half, 1)
            lava.append(((x_half - x_right_scale), z_half, x_half, -z_half))
        return lava

    def update_goal_target_image(self, goal_output):
        metadata = goal_output.metadata or {}
        # Different goal categories may use different property names
        target_names = ['target', 'targets', 'target_1', 'target_2']
        for target_name in target_names:
            # Some properties may be dicts, and some may be lists of dicts
            targets = metadata.get(target_name) or []
            targets = targets if isinstance(targets, list) else [targets]
            for target in targets:
                # Backwards compatibility: target used to have image data
                image = target.get('image')
                # Convert the image data from a string to an array
                if isinstance(image, str):
                    image_array = np.array(ast.literal_eval(image)).tolist()
                    target['image'] = image_array
        return goal_output



# Source:
# https://stackoverflow.com/questions/510357/python-read-a-single-character-from-the-user

class _Getch:
    """Gets a single character from standard input.  Does not echo to the
screen."""

    def __init__(self):
        try:
            self.impl = _GetchWindows()
        except ImportError:
            try:
                self.impl = _GetchUnix()
            except ImportError:
                self.impl = _GetchMacCarbon()

    def __call__(self):
        return self.impl()


class _GetchUnix:
    def __init__(self):
        import sys  # noqa: F401
        import tty  # noqa: F401

    def __call__(self):
        import sys
        import termios
        import tty
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch


class _GetchWindows:
    def __init__(self):
        import msvcrt  # pylint: disable=import-error # noqa: F401

    def __call__(self):
        import msvcrt  # pylint: disable=import-error # noqa: F401
        return msvcrt.getch()


class _GetchMacCarbon:
    """
    A function which returns the current ASCII key that is down;
    if no ASCII key is down, the null string is returned.  The
    page http://www.mactech.com/macintosh-c/chap02-1.html was
    very helpful in figuring out how to do this.
    """

    def __init__(self):
        import Carbon  # pylint: disable=import-error

        # pylint: disable= W0104
        Carbon.Evt  # see if it has this (in Unix, it doesn't)

    def __call__(self):
        import Carbon  # pylint: disable=import-error
        if Carbon.Evt.EventAvail(0x0008)[0] == 0:  # 0x0008 is the keyDownMask
            return ''
        #
        # The event contains the following info:
        # (what, msg, when, where, mod)=Carbon.Evt.GetNextEvent(0x0008)[1]
        #
        # The message (msg) contains the ASCII char which is
        # extracted with the 0x000000FF charCodeMask; this
        # number is converted to an ASCII character with chr() and
        # returned
        #
        (what, msg, *other) = Carbon.Evt.GetNextEvent(0x0008)[1]
        return chr(msg & 0x000000FF)


getch = _Getch()



import unittest

from machine_common_sense import LoggingConfig


class TestLoggingConfig(unittest.TestCase):
    def test_logging_config(self):
        cfg = LoggingConfig.get_configurable_logging_config(
            'INFO', ['l1', 'l2'], False, False, True, root_log_level='ERROR')
        root = cfg['root']
        loggers = cfg['loggers']
        handlers = cfg['handlers']
        self.assertEqual(root['level'], 'ERROR')
        self.assertEqual(len(loggers), 2)
        self.assertEqual(len(handlers), 1)
        self.assertIn("l1", loggers)
        self.assertIn("l2", loggers)
        self.assertEqual(loggers['l1']['level'], 'INFO')
        self.assertEqual(loggers['l2']['level'], 'INFO')
        self.assertEqual(len(loggers['l1']['handlers']), 1)
        self.assertEqual(len(loggers['l2']['handlers']), 1)
        self.assertEqual(len(root['handlers']), 1)
        self.assertEqual(loggers['l1']['handlers'], ['info-file'])
        self.assertEqual(loggers['l1']['handlers'], ['info-file'])
        self.assertEqual(root['handlers'], ['info-file'])

    def test_logging_config2(self):
        cfg = LoggingConfig.get_configurable_logging_config(
            'ERROR', 'l1', True, True, False)
        root = cfg['root']
        loggers = cfg['loggers']
        handlers = cfg['handlers']
        self.assertEqual(root['level'], 'WARN')
        self.assertEqual(len(loggers), 1)
        self.assertEqual(len(handlers), 2)
        self.assertIn("l1", loggers)
        self.assertEqual(loggers['l1']['level'], 'ERROR')
        self.assertEqual(len(loggers['l1']['handlers']), 2)
        self.assertEqual(len(root['handlers']), 2)
        self.assertIn('console', loggers['l1']['handlers'])
        self.assertIn('console', root['handlers'])
        self.assertIn('debug-file', loggers['l1']['handlers'])
        self.assertIn('debug-file', root['handlers'])


if __name__ == '__main__':
    unittest.main()



import textwrap
import unittest

import machine_common_sense as mcs


class TestGoalMetadata(unittest.TestCase):

    str_output = '''    {
        "action_list": null,
        "category": "",
        "description": "",
        "habituation_total": 0,
        "last_preview_phase_step": 0,
        "last_step": null,
        "metadata": {},
        "steps_allowed_in_lava": 0,
        "triggered_by_target_sequence": null
    }'''

    @classmethod
    def setUpClass(cls):
        cls.goal_metadata = mcs.GoalMetadata()

    @classmethod
    def tearDownClass(cls):
        pass

    def test_action_list(self):
        self.assertIsNone(self.goal_metadata.action_list)

    def test_category(self):
        self.assertEqual(self.goal_metadata.category, '')

    def test_description(self):
        self.assertEqual(self.goal_metadata.description, '')

    def test_habituation_total(self):
        self.assertEqual(self.goal_metadata.habituation_total, 0)
        self.assertIsInstance(self.goal_metadata.habituation_total, int)

    def test_last_preview_phase(self):
        self.assertEqual(self.goal_metadata.last_preview_phase_step, 0)
        self.assertIsInstance(self.goal_metadata.last_preview_phase_step, int)

    def test_last_step(self):
        self.assertIsNone(self.goal_metadata.last_step)

    def test_triggered_by_target_sequence(self):
        self.assertIsNone(self.goal_metadata.triggered_by_target_sequence)

    def test_metadata(self):
        self.assertFalse(self.goal_metadata.metadata)
        self.assertIsInstance(self.goal_metadata.metadata, dict)

    def test_retrieve_action_list_with_preview_phase(self):
        goal_metadata = mcs.GoalMetadata(
            action_list=[],
            last_preview_phase_step=10,
            last_step=10)
        self.assertEqual(
            goal_metadata.retrieve_action_list_at_step(0), [(
                'Pass', {})])
        self.assertEqual(
            goal_metadata.retrieve_action_list_at_step(9), [(
                'Pass', {})])
        self.assertEqual(self.goal_metadata.retrieve_action_list_at_step(10), [
            ('CloseObject', {}),
            ('DropObject', {}),
            ('MoveAhead', {}),
            ('MoveBack', {}),
            ('MoveLeft', {}),
            ('MoveRight', {}),
            ('OpenObject', {}),
            ('PickupObject', {}),
            ('PullObject', {}),
            ('PushObject', {}),
            ('PutObject', {}),
            ('TorqueObject', {}),
            ('RotateObject', {}),
            ('MoveObject', {}),
            ('InteractWithAgent', {}),
            ('LookUp', {}),
            ('LookDown', {}),
            ('RotateLeft', {}),
            ('RotateRight', {}),
            ('EndScene', {}),
            ('Pass', {})
        ])

    def test_retrieve_action_list_after_preview_phase(self):
        goal_metadata = mcs.GoalMetadata(
            action_list=[['MoveBack']],
            last_preview_phase_step=10,
            last_step=11)
        self.assertEqual(
            goal_metadata.retrieve_action_list_at_step(0), [(
                'Pass', {})])
        self.assertEqual(
            goal_metadata.retrieve_action_list_at_step(9), [(
                'Pass', {})])
        self.assertEqual(goal_metadata.retrieve_action_list_at_step(10), [
            ('MoveBack', {})
        ])
        self.assertEqual(goal_metadata.retrieve_action_list_at_step(11), [])

    def test_retrieve_action_list_at_final_step(self):
        goal_metadata = mcs.GoalMetadata(
            action_list=[],
            last_step=10)
        self.assertEqual(
            goal_metadata.retrieve_action_list_at_step(10), [])

    def test_retrieve_action_list_after_final_step(self):
        goal_metadata = mcs.GoalMetadata(
            action_list=[],
            last_step=10)
        self.assertEqual(
            goal_metadata.retrieve_action_list_at_step(15), [])

    def test_retrieve_action_too_many_steps_lava_default(self):
        goal_metadata = mcs.GoalMetadata(
            action_list=[],
            last_step=10)
        self.assertEqual(goal_metadata.retrieve_action_list_at_step(11, 1), [
            ('EndScene', {})
        ])

    def test_retrieve_action_incorrect_triggered_by_sequence(self):
        goal_metadata = mcs.GoalMetadata(
            action_list=[])
        self.assertEqual(goal_metadata.retrieve_action_list_at_step(
            0, triggered_by_sequence_incorrect=True), [
            ('EndScene', {})
        ])

    def test_set_triggered_by_target_sequence(self):
        goal_metadata = mcs.GoalMetadata(
            triggered_by_target_sequence=["chest_1", "chest_3"])
        self.assertEqual(
            goal_metadata.triggered_by_target_sequence, [
                "chest_1", "chest_3"])

    def test_retrieve_action_too_many_steps(self):
        goal_metadata = mcs.GoalMetadata(
            action_list=[],
            last_step=10,
            steps_allowed_in_lava=3)
        self.assertEqual(goal_metadata.retrieve_action_list_at_step(11, 1), [])
        self.assertEqual(goal_metadata.retrieve_action_list_at_step(11, 4), [
            ('EndScene', {})
        ])

    def test_retrieve_action_list_at_step(self):
        self.assertEqual(self.goal_metadata.retrieve_action_list_at_step(0), [
            ('CloseObject', {}),
            ('DropObject', {}),
            ('MoveAhead', {}),
            ('MoveBack', {}),
            ('MoveLeft', {}),
            ('MoveRight', {}),
            ('OpenObject', {}),
            ('PickupObject', {}),
            ('PullObject', {}),
            ('PushObject', {}),
            ('PutObject', {}),
            ('TorqueObject', {}),
            ('RotateObject', {}),
            ('MoveObject', {}),
            ('InteractWithAgent', {}),
            ('LookUp', {}),
            ('LookDown', {}),
            ('RotateLeft', {}),
            ('RotateRight', {}),
            ('EndScene', {}),
            ('Pass', {})
        ])
        self.assertEqual(self.goal_metadata.retrieve_action_list_at_step(10), [
            ('CloseObject', {}),
            ('DropObject', {}),
            ('MoveAhead', {}),
            ('MoveBack', {}),
            ('MoveLeft', {}),
            ('MoveRight', {}),
            ('OpenObject', {}),
            ('PickupObject', {}),
            ('PullObject', {}),
            ('PushObject', {}),
            ('PutObject', {}),
            ('TorqueObject', {}),
            ('RotateObject', {}),
            ('MoveObject', {}),
            ('InteractWithAgent', {}),
            ('LookUp', {}),
            ('LookDown', {}),
            ('RotateLeft', {}),
            ('RotateRight', {}),
            ('EndScene', {}),
            ('Pass', {})
        ])

    def test_retrieve_action_list_hidden_endhabituation_params(self):
        goal_metadata = mcs.GoalMetadata(action_list=[
            ['EndHabituation,xPosition=0,zPosition=0,yRotation=90'],
        ])
        self.assertEqual(goal_metadata.retrieve_action_list_at_step(0), [
            ('EndHabituation', {})
        ])

    def test_retrieve_action_list_at_step_with_custom_action_list(self):
        goal_metadata = mcs.GoalMetadata(action_list=[
            ['Pass'],
            ['LookDown', 'LookUp', 'RotateLeft', 'RotateRight', 'Pass'],
            [],
            ['EndHabituation,xPosition=0,zPosition=0,yRotation=90'],
            [('PickupObject', {'objectId': 'target'})]
        ])
        self.assertEqual(goal_metadata.retrieve_action_list_at_step(0), [
            ('Pass', {})
        ])
        self.assertEqual(goal_metadata.retrieve_action_list_at_step(1), [
            ('LookDown', {}),
            ('LookUp', {}),
            ('RotateLeft', {}),
            ('RotateRight', {}),
            ('Pass', {})
        ])
        self.assertEqual(goal_metadata.retrieve_action_list_at_step(2), [
            ('CloseObject', {}),
            ('DropObject', {}),
            ('MoveAhead', {}),
            ('MoveBack', {}),
            ('MoveLeft', {}),
            ('MoveRight', {}),
            ('OpenObject', {}),
            ('PickupObject', {}),
            ('PullObject', {}),
            ('PushObject', {}),
            ('PutObject', {}),
            ('TorqueObject', {}),
            ('RotateObject', {}),
            ('MoveObject', {}),
            ('InteractWithAgent', {}),
            ('LookUp', {}),
            ('LookDown', {}),
            ('RotateLeft', {}),
            ('RotateRight', {}),
            ('EndScene', {}),
            ('Pass', {})
        ])
        self.assertEqual(goal_metadata.retrieve_action_list_at_step(3), [
            ('EndHabituation', {})
        ])
        self.assertEqual(goal_metadata.retrieve_action_list_at_step(4), [
            ('PickupObject', {'objectId': 'target'})
        ])
        self.assertEqual(goal_metadata.retrieve_action_list_at_step(10), [
            ('CloseObject', {}),
            ('DropObject', {}),
            ('MoveAhead', {}),
            ('MoveBack', {}),
            ('MoveLeft', {}),
            ('MoveRight', {}),
            ('OpenObject', {}),
            ('PickupObject', {}),
            ('PullObject', {}),
            ('PushObject', {}),
            ('PutObject', {}),
            ('TorqueObject', {}),
            ('RotateObject', {}),
            ('MoveObject', {}),
            ('InteractWithAgent', {}),
            ('LookUp', {}),
            ('LookDown', {}),
            ('RotateLeft', {}),
            ('RotateRight', {}),
            ('EndScene', {}),
            ('Pass', {})
        ])

    def test_retrieve_action_list_passive_scene(self):
        goal_metadata = mcs.GoalMetadata(action_list=[], last_step=10)
        self.assertEqual(
            goal_metadata.retrieve_action_list_at_step(0, 0, False, True),
            [('Pass', {})]
        )

    def test_str(self):
        self.assertEqual(str(self.goal_metadata),
                         textwrap.dedent(self.str_output))


if __name__ == '__main__':
    unittest.main()
